"""Functions for plotting different types of data or reaction kinetics, all using matplotlib."""

import itertools
from collections.abc import Callable
from copy import deepcopy
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib.ticker import FuncFormatter

from cobrak.io import json_load


def _get_constant_combinations(
    n: int,
    variable_combination: tuple[int, int],
    min_values: list[float],
    max_values: list[float],
) -> list[list[float]]:
    """Generate all possible combinations of constant values for the other arguments.

    Args:
    - n: The total number of arguments.
    - variable_combination: A tuple of two variable argument indices.
    - min_values: A list of minimum possible values for each argument.
    - max_values: A list of maximum possible values for each argument.

    Returns:
    - A list of lists, where each sublist contains a combination of constant values.
    """
    constant_combinations = []
    constant_values = [
        min_values,
        [(max_values[i] - min_values[i]) / 2 + min_values[i] for i in range(n)],
        max_values,
    ]
    constant_indices = [i for i in range(n) if i not in variable_combination]
    for combination in itertools.product(range(3), repeat=len(constant_indices)):
        constant_combination = [0.0] * n
        for i, index in enumerate(constant_indices):
            constant_combination[index] = constant_values[combination[i]][index]
        constant_combinations.append(constant_combination)
    return constant_combinations


def dual_axis_plot(
    xpoints: list[float],
    leftaxis_ypoints_list: list[list[float]],
    rightaxis_ypoints_list: list[list[float]],
    xaxis_caption: str = "",
    leftaxis_caption: str = "",
    rightaxis_caption: str = "",
    leftaxis_colors: list[str] = [],
    rightaxis_colors: list[str] = [],
    leftaxis_titles: list[str] = [],
    rightaxis_titles: list[str] = [],
    extrapoints: list[tuple[float, float, bool, str, str, str, float]] = [],
    has_legend: bool = True,
    legend_direction: str = "",
    legend_position: tuple[Any, ...] = (),
    is_leftaxis_logarithmic: bool = False,
    is_rightaxis_logarithmic: bool = False,
    point_style: str = "",
    line_style: str = "-",
    max_digits_after_comma: int = 4,
    savepath: str = "",
    left_ylim: None | tuple[float, float] = None,
    right_ylim: None | tuple[float, float] = None,
    xlim: None | tuple[float, float] = None,
    left_axis_in_front: bool = True,
    left_legend_position: list[int] = [],
    right_legend_position: list[int] = [],
    figure_size_inches: None | tuple[float, float] = None,
    special_figure_mode: bool = False,
) -> None:
    """Creates a plot with a dual Y-axis.

    Args:
        xpoints (list[float]): X-axis data points.
        leftaxis_ypoints_list (list[list[float]]): List of Y-axis data points for the left axis.
        rightaxis_ypoints_list (list[list[float]]): List of Y-axis data points for the right axis.
        xaxis_caption (str, optional): X-axis caption. Defaults to "".
        leftaxis_caption (str, optional): Left Y-axis caption. Defaults to "".
        rightaxis_caption (str, optional): Right Y-axis caption. Defaults to "".
        leftaxis_colors (list[str], optional): Colors for left axis lines. Defaults to [].
        rightaxis_colors (list[str], optional): Colors for right axis lines. Defaults to [].
        leftaxis_titles (list[str], optional): Legend titles for left axis lines. Defaults to [].
        rightaxis_titles (list[str], optional): Legend titles for right axis lines. Defaults to [].
        extrapoints (list[tuple[float, float, bool, str, str, str, float]], optional): List of single points,
            described by tuples with the content [x, y, is_left_axis, color, marker, label, yerr]. If yerr=0,
            no error bar is drawn at all. Defaults to [].
        has_legend (bool, optional): Whether to show the legend. Defaults to True.
        legend_direction (str, optional): Legend direction. Defaults to "".
        legend_position (tuple[float, float], optional): Legend position. Defaults to ().
        is_leftaxis_logarithmic (bool, optional): Whether to use a logarithmic scale for the left axis. Defaults to False.
        is_rightaxis_logarithmic (bool, optional): Whether to use a logarithmic scale for the right axis. Defaults to False.
        point_style (str, optional): Style for points. Defaults to "".
        line_style (str, optional): Style for lines. Defaults to "-".
        max_digits_after_comma (int, optional): Max digits after comma shown. Defaults to 4.
        savepath (str): If given, the plot is not shown but saved at the given path. Defaults to ""

    Returns:
        None (displays the plot)
    """

    fig, ax1 = plt.subplots()
    if figure_size_inches is not None:
        fig.set_size_inches(figure_size_inches[0], figure_size_inches[1])

    # Left Axis Plotting
    for i, ypoints in enumerate(leftaxis_ypoints_list):
        color = leftaxis_colors[i] if leftaxis_colors else None
        title = leftaxis_titles[i] if leftaxis_titles else None
        ax1.plot(
            xpoints,
            ypoints,
            color=color,
            linestyle=line_style,
            marker=point_style,
            label=title,
        )

    ax1.set_xlabel(xaxis_caption, fontsize=14)
    ax1.set_ylabel(leftaxis_caption, fontsize=14)
    if is_leftaxis_logarithmic:
        ax1.set_yscale("log")
    if left_ylim is not None:
        ax1.set_ylim(left_ylim[0], left_ylim[1])

    plt.xticks(fontsize=13)
    plt.yticks(fontsize=13)

    # Right Axis Plotting
    if len(rightaxis_ypoints_list) > 0:
        ax2 = ax1.twinx()
        for i, ypoints in enumerate(rightaxis_ypoints_list):
            color = rightaxis_colors[i] if rightaxis_colors else None
            title = rightaxis_titles[i] if rightaxis_titles else None
            ax2.plot(
                xpoints,
                ypoints,
                color=color,
                linestyle=line_style,
                marker=point_style,
                label=title,
            )

        ax2.set_ylabel(rightaxis_caption, fontsize=14)
        if is_rightaxis_logarithmic:
            ax2.set_yscale("log")
        if right_ylim is not None:
            ax2.set_ylim(right_ylim[0], right_ylim[1])

        if left_axis_in_front:
            ax1.set_zorder(ax2.get_zorder() + 1)
            ax1.patch.set_visible(False)

    if xlim is not None:
        ax1.set_xlim(xlim[0], xlim[1])

    for i, extrapoint in enumerate(extrapoints):
        axis = ax1 if extrapoint[2] else ax2
        if extrapoint[6] != 0.0:
            axis.errorbar(
                extrapoint[0],
                extrapoint[1],
                yerr=extrapoint[6],
                ecolor=extrapoint[3],
                capsize=5,
                linestyle="",
                color=extrapoint[3],
                marker=extrapoint[4],
                label=extrapoint[5],
            )
        else:
            axis.plot(
                extrapoint[0],
                extrapoint[1],
                linestyle="",
                color=extrapoint[3],
                marker=extrapoint[4],
                label=extrapoint[5],
            )

    # Legend
    if has_legend:
        handles, labels = ax1.get_legend_handles_labels()
        if len(rightaxis_ypoints_list) > 0:
            handles2, labels2 = ax2.get_legend_handles_labels()

            if left_legend_position != []:
                oldhandles, oldlabels = deepcopy(handles), deepcopy(labels)
                for i, left_legend_position in enumerate(left_legend_position):
                    handles[left_legend_position] = oldhandles[i]
                    labels[left_legend_position] = oldlabels[i]

            if right_legend_position != []:
                oldhandles2, oldlabels2 = deepcopy(handles2), deepcopy(labels2)
                for i, right_legend_position in enumerate(right_legend_position):
                    handles2[right_legend_position] = oldhandles2[i]
                    labels2[right_legend_position] = oldlabels2[i]
            if special_figure_mode:
                # Just for COBRA-k's initial publication :-)
                del handles[1]
                del labels[1]
                handles2.append(oldhandles[-2])
                labels2.append(oldlabels[-2])

            handles = handles + handles2
            labels = labels + labels2
        extraargs = {"loc": legend_position} if legend_position != () else {}
        if legend_direction:
            extraargs["loc"] = legend_direction
        plt.legend(
            handles,
            labels,
            prop={"size": 10.5},
            bbox_to_anchor=(0.5, 0.5)
            if not legend_position and not legend_direction
            else None,
            **extraargs,
        )

    plt.xticks(fontsize=13)
    plt.yticks(fontsize=13)

    # Format axis ticks
    ax1.xaxis.set_major_formatter(
        plt.FuncFormatter(lambda x, _: f"{x:.{max_digits_after_comma}f}")
    )
    ax1.yaxis.set_major_formatter(
        plt.FuncFormatter(lambda x, _: f"{x:.{max_digits_after_comma}f}")
    )
    if len(rightaxis_ypoints_list) > 0:
        ax2.yaxis.set_major_formatter(
            plt.FuncFormatter(lambda x, _: f"{x:.{max_digits_after_comma}f}")
        )

    plt.tight_layout()  # Adjust layout to prevent labels from overlapping

    if not savepath:
        plt.show()
    else:
        plt.savefig(savepath)

    # Close the plot to free up memory
    plt.close()


def plot_combinations(
    func: Callable[[list[float]], float],
    min_values: list[float],
    max_values: list[float],
    num_subplots_per_window: int = 18,
    num_subplots_per_row: int = 6,
) -> None:
    """Plot all unique combinations of 2 variable arguments and constant values for the other arguments.

    The plot is a scatter plot with different colors for each category in the hue column. The x-axis represents the x-data,
    the y-axis represents the y-data, and the hue axis represents the category.

    The plot has the following features:

    * A title at the top of the plot with the specified title.
    * Labels for the x-axis and y-axis with the specified labels.
    * A legend on the right side of the plot with the specified hue label.
    * Different colors for each category in the hue column, specified by the palette.
    * A scatter plot with points representing the data.

    Example usage:
    >>> min_values = [-1.0, 0.0, 0.0]
    >>> max_values = [10.0, 5.0, 10.0]
    >>> def example_func(args: List[float]) -> float:
    >>>    return args[0] + args[1] + args[2]
    >>>
    >>> plot_combinations(example_func, min_values, max_values)

    Args:
    - func: The function to be plotted. It takes a list of floats and returns a float.
    - min_values: A list of minimum possible values for each argument.
    - max_values: A list of maximum possible values for each argument.
    - num_subplots_per_window: The maximum number of subplots per window. Defaults to 18.
    - num_subplots_per_row: The maximum number of subplots per row in a window. Defaults to 6.

    Returns:
    - None
    """

    # Generate all possible combinations of 2 variable arguments
    variable_combinations = []
    for i in range(len(min_values)):
        for j in range(i + 1, len(min_values)):
            variable_combinations.append((i, j))

    # Generate all unique combinations of variable and constant arguments
    combinations = []
    for variable_combination in variable_combinations:
        constant_combinations = _get_constant_combinations(
            len(min_values), variable_combination, min_values, max_values
        )
        for constant_combination in constant_combinations:
            combinations.append((variable_combination, constant_combination))

    # Plot each combination
    num_windows = int(np.ceil(len(combinations) / num_subplots_per_window))
    for window_index in range(num_windows):
        num_subplots = min(
            num_subplots_per_window,
            len(combinations) - window_index * num_subplots_per_window,
        )
        num_rows = int(np.ceil(num_subplots / num_subplots_per_row))
        _, axs = plt.subplots(
            num_rows,
            num_subplots_per_row,
            figsize=(20, 5 * num_rows),
            subplot_kw={"projection": "3d"},
        )
        if num_subplots_per_row == 1:
            axs = [[ax] for ax in axs]
        elif num_rows == 1:
            axs = [axs]
        else:
            axs = [list(axs_row) for axs_row in axs]

        z_mins = {}
        z_maxs = {}
        for combination in combinations[
            window_index * num_subplots_per_window : (window_index + 1)
            * num_subplots_per_window
        ]:
            variable_combination, constant_combination = combination
            if variable_combination not in z_mins:
                z_mins[variable_combination] = float("inf")
                z_maxs[variable_combination] = float("-inf")
            x = np.linspace(
                min_values[variable_combination[0]],
                max_values[variable_combination[0]],
                100,
            )
            y = np.linspace(
                min_values[variable_combination[1]],
                max_values[variable_combination[1]],
                100,
            )
            X, Y = np.meshgrid(x, y)
            Z = np.zeros(X.shape)
            for i in range(X.shape[0]):
                for j in range(X.shape[1]):
                    args = [
                        (
                            X[i, j]
                            if k == variable_combination[0]
                            else (
                                Y[i, j]
                                if k == variable_combination[1]
                                else constant_combination[k]
                            )
                        )
                        for k in range(len(min_values))
                    ]
                    Z[i, j] = func(args)
            z_mins[variable_combination] = min(z_mins[variable_combination], np.min(Z))  # type: ignore
            z_maxs[variable_combination] = max(z_maxs[variable_combination], np.max(Z))  # type: ignore

        for subplot_index, combination in enumerate(
            combinations[
                window_index * num_subplots_per_window : (window_index + 1)
                * num_subplots_per_window
            ]
        ):
            variable_combination, constant_combination = combination
            x = np.linspace(
                min_values[variable_combination[0]],
                max_values[variable_combination[0]],
                100,
            )
            y = np.linspace(
                min_values[variable_combination[1]],
                max_values[variable_combination[1]],
                100,
            )
            X, Y = np.meshgrid(x, y)
            Z = np.zeros(X.shape)
            for i in range(X.shape[0]):
                for j in range(X.shape[1]):
                    args = [
                        (
                            X[i, j]
                            if k == variable_combination[0]
                            else (
                                Y[i, j]
                                if k == variable_combination[1]
                                else constant_combination[k]
                            )
                        )
                        for k in range(len(min_values))
                    ]
                    Z[i, j] = func(args)

            row_index = subplot_index // num_subplots_per_row
            col_index = subplot_index % num_subplots_per_row

            axs[row_index][col_index].plot_surface(
                X, Y, Z, cmap="viridis", edgecolor="none"
            )
            axs[row_index][col_index].set_xlabel(f"Argument {variable_combination[0]}")
            axs[row_index][col_index].set_ylabel(f"Argument {variable_combination[1]}")
            axs[row_index][col_index].set_zlim(
                z_mins[variable_combination], z_maxs[variable_combination]
            )

            constant_title = ", ".join(
                [
                    f"{i}: {constant_combination[i]}"
                    for i in range(len(min_values))
                    if i not in variable_combination
                ]
            )
            axs[row_index][col_index].set_title(
                (
                    f"Variable: {variable_combination[0]}, {variable_combination[1]}\n"
                    f"Constant: {constant_title}"
                    if constant_title
                    else f"Variable: {variable_combination[0]}, {variable_combination[1]}"
                ),
                fontsize=8,
            )

            # Hide empty plots
            if subplot_index >= num_subplots:
                axs[row_index][col_index].axis("off")

        plt.tight_layout()
        plt.show()


def plot_objvalue_evolution(
    json_path: str,
    output_path: str,
    ylabel: str = "Objective value",
    objvalue_multiplicator: float = -1.0,
    with_legend: bool = False,
    algorithm: Literal["pso", "genetic"] = "genetic",
    precision: int = 4,
) -> None:
    """Plots the evolution of the objective value over computational time.

    Args:
        json_path (str): Path to the JSON file containing the data.
        output_path (str): Path to save the plot.
        ylabel (str, optional): Label for the Y-axis. Defaults to "Objective value".
        objvalue_multiplicator (float, optional): Multiplier to apply to the objective value. Defaults to -1.0.
        with_legend (bool, optional): Whether to display the legend. Defaults to False.
        algorithm (Literal["pso", "genetic"], optional): The optimization algorithm used. Defaults to "genetic".
        precision (int, optional): The number of decimal places to display on the Y-axis. Defaults to 4.

    Returns:
        None. Saves the plot to the specified output path.
    """

    def format_decimal(x, _) -> str:  # noqa: ANN001
        return f"{x:.{precision}f}"  # Use the specified precision

    # Load data from JSON file
    data = json_load(json_path, Any)

    # Extract timepoints
    timepoints = tuple(float(key) for key in data)

    # Initialize objvalues list
    match algorithm:
        case "pso":
            objvalues = [[] for _ in range(len(list(data.values())[0]))]
        case "genetic":
            objvalues = [[]]
        case _:
            raise ValueError

    # Populate objvalues list
    for values in data.values():
        match algorithm:
            case "pso":
                for value_idx, value in enumerate(values):
                    if value >= 1_000_000.0:
                        objvalues[value_idx].append(None)
                    else:
                        objvalues[value_idx].append(value * objvalue_multiplicator)
            case "genetic":
                objvalues[0].append(objvalue_multiplicator * values[0])

    plt.clf()
    plt.cla()
    match algorithm:
        case "pso":
            for objvalue_idx, objvalue_list in enumerate(objvalues):
                plt.plot(
                    timepoints,
                    objvalue_list,
                    linestyle="-",
                    marker=None,
                    label=f"Particle {objvalue_idx + 1}",
                )
        case "genetic":
            plt.plot(
                timepoints, objvalues[0], linestyle="-", marker=None, label="Best value"
            )

    # Customize the plot
    plt.xlabel("Computational Time [s]")
    plt.ylabel(ylabel)
    plt.title(f"{ylabel} Evolution Over Time")
    if with_legend:
        plt.legend()

    plt.gca().yaxis.set_major_formatter(FuncFormatter(format_decimal))

    # Save the plot
    plt.savefig(output_path)

    # Close the plot to free up memory
    plt.close()


def plot_variabilities(
    variabilities: list[list[tuple[float, float, float]]],
    variability_names: list[str],
    variability_titles: list[str],
    colors: list[str],
    xlabel: str = "",
    ylabel: str = "",
    yscale: str = "log",
    plot_mean: bool = True,
    save_path: str | None = None,
) -> None:
    """Plots the mean values and whisker bars for multiple variabilities.

    This function generates a plot where each variability is represented by a series of points (triangles) and whisker bars.
    Each point (if plot_mean==True) represents the mean value of a data point in the variability, and the whisker bars represent the lower and upper bounds.
    The variabilities are grouped together for each data point, with a space between each group to clearly distinguish them.

    Parameters:
    -----------
    variabilities : List[List[Tuple[float, float, float]]]
    A list of lists, where each inner list represents a variability. Each tuple in the inner list contains
    (lower_bound, upper_bound, mean_value) for each data point in the variability.

    variability_names : List[str]
    A list of strings representing the names of the variabilities.

    colors : List[str]
    A list of strings representing the colors for each variability, e.g. using names from https://matplotlib.org/stable/gallery/color/named_colors.html

    plot_mean : bool, optional
    If True, the mean value is plotted as a triangle. If False, only the whisker bars are plotted. Default is True.

    save_path : str, optional
    The file path where the plot should be saved. If None, the plot is displayed. Default is None.

    Returns:
    --------
    None
    The function either displays the plot or saves it to the specified path.

    Example:
    --------
    >>> in_vivo = [(1.0, 3.0, 2.0), (2.0, 4.0, 3.0), (3.0, 5.0, 4.0)]
    >>> in_silico = [(1.5, 3.5, 2.5), (2.5, 4.5, 3.5), (3.5, 5.5, 4.5)]
    >>> another_variability = [(1.2, 3.2, 2.2), (2.2, 4.2, 3.2), (3.2, 5.2, 4.2)]
    >>> variabilities = [in_vivo, in_silico, another_variability]
    >>> variability_names = ['in_vivo', 'in_silico', 'another_variability']
    >>> colors = ['blue', 'orange', 'green']
    >>> plot_variabilities(variabilities, variability_names, colors)
    >>> plot_variabilities(variabilities, variability_names, colors, plot_mean=False)
    >>> plot_variabilities(variabilities, variability_names, colors, save_path='plot.png')
    """
    # Number of variabilities
    n = len(variabilities[0])
    num_variabilities = len(variabilities)

    # Create a figure and axis
    _, ax = plt.subplots()

    # Define the positions for the groups
    positions = [
        list(
            range(
                i * (num_variabilities + 1),
                i * (num_variabilities + 1) + num_variabilities,
            )
        )
        for i in range(n)
    ]

    # Plot each variability
    for i, (pos_group, variability) in enumerate(zip(positions, zip(*variabilities))):
        for j, (pos, (lower, upper, mean)) in enumerate(zip(pos_group, variability)):
            if plot_mean:
                ax.errorbar(
                    pos,
                    mean,
                    yerr=[[mean - lower], [upper - mean]],
                    fmt="o",
                    capsize=5,
                    color=colors[j],
                    ecolor=colors[j],
                    label=variability_titles[j] if i == 0 else "",
                )
            else:
                ax.errorbar(
                    pos,
                    mean,
                    yerr=[[mean - lower], [upper - mean]],
                    fmt="none",
                    capsize=5,
                    ecolor=colors[j],
                    label=variability_titles[j] if i == 0 else "",
                )

    # Calculate midpoints between groups for vertical lines
    for i in range(len(positions) - 1):
        # Get the end of the current group and the start of the next group
        current_group_end = positions[i][-1]
        next_group_start = positions[i + 1][0]
        # Calculate the midpoint
        midpoint = (current_group_end + next_group_start) / 2
        # Draw a thin vertical black line at the midpoint
        ax.axvline(x=midpoint, color="black", linestyle="-", linewidth=0.5, alpha=0.7)

    # Set the x-axis labels
    ax.set_xticks([pos[0] + (num_variabilities - 1) / 2 for pos in positions])
    ax.set_xticklabels(variability_names)  # [f"Exp {i+1}" for i in range(n)])

    # Add labels and title
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title("Comparison of Variabilities")
    ax.set_yscale(yscale)

    # Add legend
    ax.legend()

    # Save or show the plot
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()


def scatterplot_with_labels(
    x_data: list[tuple[float, float, float]],
    y_data: list[tuple[float, float, float]],
    labels: list[str],
    x_label: str = None,
    y_label: str = None,
    y_log: bool = True,
    x_log: bool = True,
    add_labels: bool = False,
    identical_axis_lims: bool = True,
    xlim_overwrite: None | tuple[float, float] = None,
    ylim_overwrite: None | tuple[float, float] = None,
    ax: plt.Axes = None,
    save_path: str = None,
    title: str | None = None,
    extratext: str | None = None,
    x_labelsize: float = 14,
    y_labelsize: float = 14,
) -> plt.Axes:
    """
    Generates a scatter plot with error bars and optional point labels.

    Can be used standalone ("one-off" plot with plt.show()), or for subplotting by passing an Axes object.
    Optionally saves the figure if save_path is provided.

    Parameters
    ----------
    x_data : list[tuple[float, float, float]]
        Each tuple is (lower bound, upper bound, drawn value) for x.
    y_data : list[tuple[float, float, float]]
        Each tuple is (lower bound, upper bound, drawn value) for y.
    labels : list[str]
        Labels for each point (used if add_labels is True).
    x_label : str, optional
        X-axis label.
    y_label : str, optional
        Y-axis label.
    y_log : bool, default True
        Use log scale for y-axis.
    x_log : bool, default True
        Use log scale for x-axis.
    add_labels : bool, default False
        Annotate points with corresponding label.
    identical_axis_lims : bool, default True
        Make x and y axis limits identical and auto-scale them.
    ax : matplotlib.axes.Axes, optional
        If provided, plot is drawn on this Axes (for subplotting).
    save_path : str, optional
        If provided and `ax` is None (standalone plotting), save the figure at this path instead of showing.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axis object containing the plot.
    """
    # Calculate midpoints and error sizes for x and y coordinates
    x_drawn = [x[2] for x in x_data]
    x_low = [x[0] for x in x_data]
    x_high = [x[1] for x in x_data]
    x_err_low = [x_drawn[i] - x_low[i] for i in range(len(x_data))]
    x_err_high = [x_high[i] - x_drawn[i] for i in range(len(x_data))]

    y_drawn = [y[2] for y in y_data]
    y_low = [y[0] for y in y_data]
    y_high = [y[1] for y in y_data]
    y_err_low = [y_drawn[i] - y_low[i] for i in range(len(y_data))]
    y_err_high = [y_high[i] - y_drawn[i] for i in range(len(y_data))]

    n_points = len(x_drawn)
    colors = get_cmap("viridis")(np.linspace(0, 1, n_points))

    _created_fig = False
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 6))
        _created_fig = True

    # Plot each point individually to assign different colors
    for i in range(n_points):
        ax.errorbar(
            x_drawn[i],
            y_drawn[i],
            xerr=[[x_err_low[i]], [x_err_high[i]]],
            yerr=[[y_err_low[i]], [y_err_high[i]]],
            fmt="o",
            markersize=8,
            color=colors[i],
            capsize=5,
            capthick=2,
            elinewidth=2,
        )

    # Add labels to each point
    if add_labels:
        for i, (xi, yi) in enumerate(zip(x_drawn, y_drawn)):
            ax.annotate(
                labels[i],
                (xi, yi),
                textcoords="offset points",
                xytext=(0, 10),
                ha="center",
            )

    # Axis limits & unity line
    all_x_values = [x_datapoint[0] for x_datapoint in x_data] + [
        x_datapoint[1] for x_datapoint in x_data
    ]
    all_y_values = [y_datapoint[0] for y_datapoint in y_data] + [
        y_datapoint[1] for y_datapoint in y_data
    ]

    min_val = min(*all_y_values, *all_x_values) * 0.99
    max_val = max(*all_y_values, *all_x_values) * 1.2

    if identical_axis_lims:
        ax.set_xlim(min_val, max_val)
        ax.set_ylim(min_val, max_val)

    if xlim_overwrite is not None:
        ax.set_xlim(xlim_overwrite[0], xlim_overwrite[1])
    if ylim_overwrite is not None:
        ax.set_ylim(ylim_overwrite[0], ylim_overwrite[1])

    x_unity = np.linspace(0, max_val * 100, 10)
    y_unity = x_unity
    ax.plot(x_unity, y_unity, "-", color="black", linewidth=1)

    if y_log:
        ax.set_yscale("log")
    if x_log:
        ax.set_xscale("log")

    if x_label:
        ax.set_xlabel(x_label, fontsize=x_labelsize)
    if y_label:
        ax.set_ylabel(y_label, fontsize=y_labelsize)

    if title is not None:
        ax.set_title(title, loc="left", fontweight="bold", fontsize=16)

    if extratext:
        ax.text(
            0.15,
            0.975,
            extratext,
            horizontalalignment="center",
            verticalalignment="top",
            transform=ax.transAxes,
            fontsize=14,
            fontweight="bold",
        )

    ax.grid(True)

    ax.tick_params(axis="both", which="major", labelsize=13)
    ax.tick_params(axis="both", which="minor", labelsize=10)

    if _created_fig:
        plt.tight_layout()
        if save_path is not None:
            plt.savefig(save_path)
        else:
            plt.show()
        plt.close(fig)
    return ax



"""
COBRAk is a general COBRA/COBRA-k suite written as Python package. For more about it, visit its repository:
https://github.com/klamt-lab/COBRAk

The __init__ file of COBRAk initializes the rich text output & tracebacks as well as its logger. Furthermore, graceful
shutdown of user-induced shutdowns is enabled.
"""

import logging
import signal
import sys
from types import FrameType
from typing import Any, Optional

from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import install


# PUBLIC FUNCTIONS SECTION #
def exit_signal_handler(sig: int, frame: FrameType | None) -> None:
    """Handles the exit signal by printing a shutdown message and exiting the program.

    Args:
        sig (int): The signal number.
        frame (Optional[FrameType]): The current frame.
    """
    print(
        "COBRAk received user signal to terminate (this message may appear multiple times in parallelized contexts). Shutting down..."
    )
    sys.exit(0)


def setup_rich_tracebacks(show_locals: bool) -> None:
    """Sets up rich tracebacks with the given options.

    Args:
        show_locals (bool): Whether to show local variables in the traceback.
    """
    install(show_locals=show_locals)


def set_logging_level(level: int) -> None:
    """Sets the logging level.

    E.g. INFO, ERROR, WARNING and CRITICAL from Python's logging module.

    Args:
        level (int): The logging level.
    """
    logger.setLevel(level)


def set_logging_handler(
    show_path: bool = False,
    show_time: bool = False,
    show_level: bool = True,
    keywords: list[str] = [
        "info",
        "warning",
        "error",
        "critical",
    ],
    **args: Any,
) -> RichHandler:
    """
    Sets up the logging handler with the given options.

    Args:
        show_path (bool, optional): Whether to show the path. Defaults to False.
        show_time (bool, optional): Whether to show the time. Defaults to False.
        show_level (bool, optional): Whether to show the level. Defaults to True.
        keywords (Dict[str, str], optional): The keywords to highlight. Defaults to ["info", "warning", "error", "critical"]
        **args (Any, optional): Additional Rich handler arguments.
    """
    return RichHandler(
        show_path=show_path,
        show_time=show_time,
        show_level=show_level,
        keywords=keywords,
        **args,
    )


# ACTIVE SCRIPT SECTION #
signal.signal(signal.SIGINT, exit_signal_handler)
signal.signal(signal.SIGTERM, exit_signal_handler)

setup_rich_tracebacks(False)

# Create a logger
logger = logging.getLogger(__name__)

# Set the logging level
set_logging_level(logging.INFO)

# Add the handler to the logger
logger.addHandler(set_logging_handler())

console = Console()



from cobrak.cobrapy_model_functionality import (
    create_irreversible_cobrapy_model_from_stoichiometries,
)


def test_create_irreversible_cobrapy_model_from_stoichiometries():
    model = create_irreversible_cobrapy_model_from_stoichiometries(
        {
            "EX_A": {"A": +1},
            "R1": {"A": -1, "B": +1},
            "EX_B": {"B": -1},
        }
    )
    assert len(model.reactions) == 3
    assert len(model.metabolites) == 2



from cobrak.ncbi_taxonomy_functionality import (
    get_taxonomy_scores,
    most_taxonomic_similar,
)


def test_get_taxonomy_scores(tmp_path):
    base_species = "Escherichia coli"
    taxonomy_dict = {
        "Escherichia coli": ["Escherichia", "Bacteria", "Organism"],
        "Pseudomonas": ["Pseudomonas", "Bacteria", "Organism"],
        "Homo sapiens": ["Homo", "Mammalia", "Animalia", "Organism"],
    }
    taxonomy_scores = get_taxonomy_scores(base_species, taxonomy_dict)
    assert taxonomy_scores["Escherichia coli"] == 0
    assert taxonomy_scores["Pseudomonas"] == 1
    assert taxonomy_scores["Homo sapiens"] == 2


def test_most_taxonomic_similar(tmp_path):
    base_species = "Escherichia coli"
    taxonomy_dict = {
        "Escherichia coli": ["Escherichia", "Bacteria", "Organism"],
        "Pseudomonas": ["Pseudomonas", "Bacteria", "Organism"],
        "Homo sapiens": ["Homo", "Mammalia", "Animalia", "Organism"],
    }
    most_similar = most_taxonomic_similar(base_species, taxonomy_dict)
    assert most_similar["Escherichia coli"] == 0
    assert most_similar["Pseudomonas"] == 1
    assert most_similar["Homo sapiens"] == 2



"""bigg_parse_metabolites_file.py

This module contains a function which transforms a BIGG metabolites .txt list
into an machine-readable JSON.
"""

# IMPORTS SECTION #
from .io import json_write


# PUBLIC FUNCTIONS SECTION #
def bigg_parse_metabolites_file(
    bigg_metabolites_txt_path: str,
    bigg_metabolites_json_path: str,
) -> None:
    """Parses a BIGG metabolites text file and returns a dictionary for this file.

    As of Sep 14 2024, a BIGG metabolites list of all BIGG-included metabolites
    is retrievable under http://bigg.ucsd.edu/data_access

    Arguments
    ----------
    * bigg_metabolites_file_path: str ~ The file path to the BIGG metabolites file.
      The usual file name (which has to be included too in this argument) is
      bigg_models_metabolites.txt
    * output_folder: str ~ The folder in which the JSON including the parsed BIGG
      metabolites file data is stored with the name 'bigg_id_name_mapping.json'

    Output
    ----------
    * A JSON file with the name 'bigg_id_name_mapping.json' in the given output folder,
      with the following structure:
    <pre>
     {
         "$BIGG_ID": "$CHEMICAL_OR_USUAL_NAME",
         (...),
         "$BIGG_ID": "$BIGG_ID",
         (...),
     }
    </pre>
    The BIGG ID <-> BIGG ID mapping is done for models which already use the BIGG IDs.
    """
    # Open the BIGG metabolites file as string list, and remove all newlines
    with open(bigg_metabolites_txt_path, encoding="utf-8") as f:
        lines = f.readlines()
    lines = [x.replace("\n", "") for x in lines if len(x) > 0]

    # Mapping variable which will store the BIGG ID<->
    bigg_id_name_mapping = {}
    # Go through each BIGG metabolites file line (which is a tab-separated file)
    # and retrieve the BIGG ID and the name (if there is a name for the given BIGG
    # ID)
    for line in lines:
        bigg_id = line.split("\t")[1]
        bigg_id_name_mapping[bigg_id] = bigg_id

        # Exception to check if there is no name :O
        try:
            name = line.split("\t")[2].lower()
        except Exception:
            continue
        bigg_id_name_mapping[name] = bigg_id

        try:
            database_links = line.split("\t")[4]
        except Exception:
            continue
        for database_link_part in database_links.split(": "):
            if "CHEBI:" not in database_link_part:
                continue
            subpart = database_link_part.split("CHEBI:")[1].strip()
            chebi_id = subpart.split("; ")[0] if "; " in subpart else subpart
            bigg_id_name_mapping[chebi_id] = bigg_id

    # Write the JSON in the given folder :D
    json_write(bigg_metabolites_json_path, bigg_id_name_mapping)



from cobrak.constants import LNCONC_VAR_PREFIX
from cobrak.dataclasses import Metabolite, Model, Reaction
from cobrak.tellurium_functionality import (
    _get_numbersafe_id,
    _get_reaction_string_of_cobrak_reaction,
    get_tellurium_string_from_cobrak_model,
)


def test_get_reaction_string():
    cobrak_model = Model(
        reactions={
            "reaction_id": Reaction(
                min_flux=0.0,
                max_flux=1.0,
                stoichiometries={"metabolite_id": 1.0},
                dG0=1.0,
                dG0_uncertainty=0.1,
                enzyme_reaction_data=None,
            ),
        },
        metabolites={"metabolite_id": Metabolite()},
        enzymes={},
        max_prot_pool=1.0,
        extra_linear_constraints=[],
        kinetic_ignored_metabolites=[],
        R=1.0,
        T=1.0,
    )
    reac_id = "reaction_id"
    cobrak_reaction = cobrak_model.reactions[reac_id]
    e_conc = 1.0
    met_concs = {"metabolite_id": 1.0}
    reac_flux = 1.0
    nlp_results = {LNCONC_VAR_PREFIX + "metabolite_id": 0.0}
    kinetic_ignored_metabolites = []
    unoptimized_reactions = {}
    reac_string = _get_reaction_string_of_cobrak_reaction(
        cobrak_model,
        reac_id,
        cobrak_reaction,
        e_conc,
        met_concs,
        reac_flux,
        nlp_results,
        kinetic_ignored_metabolites,
        unoptimized_reactions,
    )
    assert isinstance(reac_string, str)


def test_get_tellurium_string():
    cobrak_model = Model(
        reactions={
            "reaction_id": Reaction(
                min_flux=0.0,
                max_flux=1.0,
                stoichiometries={"metabolite_id": 1.0},
                dG0=1.0,
                dG0_uncertainty=0.1,
                enzyme_reaction_data=None,
            ),
        },
        metabolites={"metabolite_id": Metabolite()},
        enzymes={},
        max_prot_pool=1.0,
        extra_linear_constraints=[],
        kinetic_ignored_metabolites=[],
        R=1.0,
        T=1.0,
    )
    cell_density = 1.0
    e_concs = {"reaction_id": 1.0}
    met_concs = {"metabolite_id": 1.0}
    nlp_results = {"reaction_id": 1.0, LNCONC_VAR_PREFIX + "metabolite_id": 1e-6}
    tellurium_string = get_tellurium_string_from_cobrak_model(
        cobrak_model,
        cell_density,
        e_concs,
        met_concs,
        nlp_results,
    )
    assert isinstance(tellurium_string, str)


def test_get_numbersafe_id():
    met_id = "metabolite_id"
    numbersafe_id = _get_numbersafe_id(met_id)
    assert numbersafe_id == met_id

    met_id = "1metabolite_id"
    numbersafe_id = _get_numbersafe_id(met_id)
    assert numbersafe_id == "x1metabolite_id"


if __name__ == "__main__":
    test_get_reaction_string()
    test_get_tellurium_string()
    test_get_numbersafe_id()
    print("All tests passed.")



from setuptools import setup

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="cobrak",
    description="Implementation of COBRA-k",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/asciidoc",
    packages=["cobrak"],
    package_dir={"cobrak": "cobrak"},
    # package_data={'cobrak': [
    #     'data/*.svg',
    # ]}
)



"""Functions and associated dataclasses for retrieving kinetic data from SABIO-RK"""

# IMPORT SECTION #
import threading
from dataclasses import dataclass
from io import StringIO
from math import sqrt
from os.path import exists
from shutil import rmtree
from statistics import median
from time import sleep
from zipfile import ZIP_LZMA, ZipFile

import cobra
import requests
from dataclasses_json import dataclass_json

from .dataclasses import EnzymeReactionData, ParameterReference
from .io import (
    ensure_folder_existence,
    get_files,
    json_load,
    json_zip_load,
    standardize_folder,
)
from .ncbi_taxonomy_functionality import (
    get_taxonomy_dict_from_nbci_taxonomy,
    get_taxonomy_scores,
)


# DATACLASSES SECTION #
@dataclass_json
@dataclass
class SabioEntry:
    """Represents the COBRAk-relevant data retrieved from a single SABIO-RK entry.

    Of which type this entry is (k_cat, k_m, k_i) is not determined here. This is done
    in the dataclass SabioDict.
    """

    entry_id: int
    """The entry's ID number"""
    is_recombinant: bool
    """Whether or not the entry is from a recombinant enzyme"""
    kinetics_mechanism_type: str
    """The reaction's kinetic mechanism (e.g., "Michaelis-Menten")"""
    organism: str
    """The organism (latin-greek name) associated with this entry"""
    temperature: float | None
    """[None if not given] The measurement's temperature in °C"""
    ph: float | None
    """[None if not given] The measurement's pH"""
    parameter_value: float
    """The value of the parameter"""
    parameter_unit: str
    """The unit of the value"""
    parameter_associated_species: str
    """The species (metabolite) associated with the parameter"""
    substrates: list[str]
    """The list of substrate names"""
    products: list[str]
    """The list of product names"""
    chebi_ids: list[str]
    """The list of all CHEBI IDs"""


@dataclass_json
@dataclass
class SabioDict:
    """Includes all retrieved SabioEntry instances and shows of which type they are"""

    kcat_entries: dict[str, list[SabioEntry]]
    """Turnover number entries"""
    km_entries: dict[str, list[SabioEntry]]
    """Michaelis-Menten constant entries"""
    ki_entires: dict[str, list[SabioEntry]]
    """Inhibition constant entries"""
    ka_entires: dict[str, list[SabioEntry]]
    """Activation constant entries"""
    hill_entires: dict[str, list[SabioEntry]]
    """Hill number entries"""


# CLASS SECTION #
class SabioThread(threading.Thread):
    """Represents a single Sabio-RK connection, ready for multi-threading (on one CPU core) using the threading module"""

    def __init__(self, temp_folder: str, start_number: int, end_number: int) -> None:
        """Initializes a SabioThread instance.

        Args:
            temp_folder (str): The path to the temporary folder where the results will be saved.
            start_number (int): The starting number for the query range.
            end_number (int): The ending number for the query range.
        """
        super().__init__()

        self.temp_folder = standardize_folder(temp_folder)
        self.start_number = start_number
        self.end_number = end_number

    def run(self) -> None:
        """Executes the thread's SABIO-RK data request

        Constructs a query string, sends a POST request to the SABIO-RK web service,
        and writes the response to a file in the temporary folder.
        """
        query_numbers = " OR ".join(
            [str(i + 1) for i in range(self.start_number, self.end_number + 1)]
        )
        query_dict = {"EntryID": f"({query_numbers})"}
        query_string = " AND ".join([f"{k}:{v}" for k, v in query_dict.items()])
        query_string += ' AND Parametertype:("activation constant" OR "Ki" OR "kcat" OR "km" OR "Hill coefficient") AND EnzymeType:"wildtype"'
        query = {
            "fields[]": [
                "EntryID",
                "Organism",
                "IsRecombinant",
                "ECNumber",
                "KineticMechanismType",
                "SabioCompoundID",
                "ChebiID",
                "Parameter",
                "Substrate",
                "Product",
                "Temperature",
                "pH",
            ],
            "q": query_string,
        }
        request = requests.post(
            "http://sabiork.h-its.org/sabioRestWebServices/kineticlawsExportTsv",
            params=query,
            timeout=1e6,
        )
        request.raise_for_status()
        with open(
            f"{self.temp_folder}zzz{self.start_number}.txt", "w", encoding="utf-8"
        ) as f:
            f.write(request.text)


# "PRIVATE" FUNCTIONS SECTION #
def _get_ec_code_entries(
    sabio_entries: dict[str, list[SabioEntry]],
    reac_ec_codes: list[str],
    min_ph: float,
    max_ph: float,
    accept_nan_ph: bool,
    min_temperature: float,
    max_temperature: float,
    accept_nan_temperature: bool,
    substrate_bigg_ids: list[str],
    product_bigg_ids: list[str],
    name_to_bigg_id_dict: dict[str, str],
) -> list[SabioEntry]:
    """Filters SABIO-RK entries based on EC codes, pH, temperature, and substrate/product BIGG IDs.

    Args:
        sabio_entries (dict[str, list[SabioEntry]]): Dictionary of SabioEntry instances keyed by EC code.
        reac_ec_codes (list[str]): List of reaction EC codes to filter by.
        min_ph (float): Minimum pH value for filtering.
        max_ph (float): Maximum pH value for filtering.
        accept_nan_ph (bool): Whether to accept entries with NaN pH values.
        min_temperature (float): Minimum temperature value for filtering (in °C).
        max_temperature (float): Maximum temperature value for filtering  (in °C).
        accept_nan_temperature (bool): Whether to accept entries with NaN temperature values.
        substrate_bigg_ids (list[str]): List of substrate BIGG IDs to filter by.
        product_bigg_ids (list[str]): List of product BIGG IDs to filter by.
        name_to_bigg_id_dict (dict[str, str]): Dictionary mapping compound names to BIGG IDs.

    Returns:
        list[SabioEntry]: List of filtered SabioEntry instances.
    """
    ec_code_entries = []
    for ec_code, sabio_entry_list in sabio_entries.items():
        for sabio_entry in sabio_entry_list:
            if ec_code not in reac_ec_codes:
                continue

            # Check temperature
            if not accept_nan_temperature:  # noqa: SIM102
                if (
                    sabio_entry.temperature is None
                    or sabio_entry.temperature > max_temperature
                    or sabio_entry.temperature < min_temperature
                ):
                    continue

            # Check pH
            if not accept_nan_ph:  # noqa: SIM102
                if (
                    sabio_entry.ph is None
                    or sabio_entry.ph > max_ph
                    or sabio_entry.ph < min_ph
                ):
                    continue

            # Check correction
            correct_direction = False
            for substrate in sabio_entry.substrates:
                substrate_bigg_id = _search_metname_in_bigg_ids(
                    met_id=substrate.lower(),
                    bigg_id="",
                    entry=sabio_entry,
                    name_to_bigg_id_dict=name_to_bigg_id_dict,
                )
                if substrate_bigg_id in substrate_bigg_ids:
                    correct_direction = True
                    break
            if not correct_direction:
                for product in sabio_entry.products:
                    product_bigg_id = _search_metname_in_bigg_ids(
                        met_id=product.lower(),
                        bigg_id="",
                        entry=sabio_entry,
                        name_to_bigg_id_dict=name_to_bigg_id_dict,
                    )
                    if product_bigg_id in product_bigg_ids:
                        correct_direction = True
                        break
            if not correct_direction:
                continue

            # If all checks ok: Add this entry :-)
            ec_code_entries.append(sabio_entry)
    return ec_code_entries


def _get_sabio_tsv_str(target_folder: str) -> str:
    """Retrieves the SABIO-RK TSV string from the target folder.

    If the a zipped TSV cache does not exist, it downloads the data in threaded SabioThread instances,
    processes it, and stores it in a zip file cache.

    Args:
        target_folder (str): The path to the folder where the TSV file or zip file should be stored.

    Returns:
        str: The content of the SABIO-RK TSV file as a string.

    Example:
        target_folder = "/path/to/target/folder"
        tsv_str = _get_sabio_tsv_str(target_folder)
        print(tsv_str)
    """
    target_folder = standardize_folder(target_folder)
    ensure_folder_existence(target_folder)
    zip_filename = f"{target_folder}sabio_single_tsvs.zip"
    tsv_zip_filename = f"{target_folder}sabio_full_tsv.zip"
    sabio_tsv_filename = "sabio.tsv"

    starts = [0 + 250 * i for i in range(80_000 // 250)]

    if not exists(zip_filename):
        print("READING OUT SABIO-RK ONLINE...")

        temp_folder = f"{target_folder}sabiotemp/"
        ensure_folder_existence(temp_folder)

        start_and_end_numbers = [(start, start + 249) for start in starts]

        threads = [
            SabioThread(temp_folder, start_number, end_number)
            for start_number, end_number in start_and_end_numbers
        ]
        for thread in threads:
            thread.start()
            sleep(2.0)
        for thread in threads:
            thread.join()

        tsv_filenames = get_files(temp_folder)
        with ZipFile(zip_filename, "w", ZIP_LZMA) as zipf:
            for filename in tsv_filenames:
                zipf.write(temp_folder + filename, arcname=filename)

        rmtree(temp_folder)

    if not exists(tsv_zip_filename):
        full_tsv_lines = []
        first = True
        with ZipFile(zip_filename, "r") as zipf:
            file_names = zipf.namelist()
            for file_name in file_names:
                with zipf.open(file_name) as file:
                    lines = file.readlines()
                    for line in lines:
                        decoded_line = line.decode("utf-8")
                        if (not first) and ("EntryID	Organism" in decoded_line):
                            continue
                        if len(decoded_line) <= 1:
                            continue
                        full_tsv_lines.append(decoded_line.replace("\n", ""))
                first = False

        with ZipFile(tsv_zip_filename, "w", ZIP_LZMA) as zipf:
            zipf.writestr(
                sabio_tsv_filename, StringIO("\n".join(full_tsv_lines)).getvalue()
            )

    with ZipFile(tsv_zip_filename, "r") as zipf:  # noqa: SIM117
        with zipf.open(sabio_tsv_filename) as file:
            tsv_content = file.read().decode("utf-8")

    return tsv_content.replace("\r", "")


def _search_metname_in_bigg_ids(
    met_id: str,
    bigg_id: str,
    entry: SabioEntry,
    name_to_bigg_id_dict: dict[str, str],
) -> str:
    entry_bigg_id: str = ""
    for missing_addition_end in ["", "1", "__L", "__D"]:
        for missing_addition_front in ["", "alpha-", "beta-", "l-", "d-"]:
            addition_name = f"{missing_addition_front}{met_id}{missing_addition_end}"
            if addition_name in name_to_bigg_id_dict:
                entry_bigg_id = name_to_bigg_id_dict[addition_name]
                break
    if bigg_id:
        for chebi_id in entry.chebi_ids:
            if chebi_id in name_to_bigg_id_dict:
                entry_bigg_id = name_to_bigg_id_dict[chebi_id]
                if entry_bigg_id == bigg_id:
                    break
    return entry_bigg_id


# "PUBLIC" FUNCTIONS SECTION #
def get_full_sabio_dict(sabio_target_folder: str) -> SabioDict:
    """Parses a SABIO-RK web query TSV file from the target folder to create a SabioDict instance containing SABIO-RK entries.

    Args:
        sabio_target_folder (str): The path to the folder containing the TSV file.

    Returns:
        SabioDict: A SabioDict instance whichm in turn, contains SabioEntry instances
    """
    tsv_str = _get_sabio_tsv_str(sabio_target_folder)

    tsv_lines = tsv_str.split("\n")
    titles = tsv_lines[0].split("\t")
    del tsv_lines[0]

    sabio_dict = SabioDict({}, {}, {}, {}, {})
    for tsv_line in tsv_lines:
        line = tsv_line.split("\t")

        parameter_value_str = line[titles.index("parameter.startValue")]
        if not parameter_value_str:
            continue
        parameter_value = float(parameter_value_str)
        if parameter_value <= 0.0:
            continue  # There is no kinetic parameter that is just 0 or below

        parameter_type_str = line[titles.index("parameter.type")]
        match parameter_type_str.lower():
            case "kcat":
                sabio_dict_pointer = sabio_dict.kcat_entries
            case "km":
                sabio_dict_pointer = sabio_dict.km_entries
            case "ki":
                sabio_dict_pointer = sabio_dict.ki_entires
            case "activation constant":
                sabio_dict_pointer = sabio_dict.ka_entires
            case "hill coefficient":
                sabio_dict_pointer = sabio_dict.hill_entires
            case _:
                continue

        ec_number = line[titles.index("ECNumber")]
        entry_id = int(line[titles.index("EntryID")])
        organism = line[titles.index("Organism")]
        is_recombinant = line[titles.index("IsRecombinant")].lower() == "true"
        kinetics_mechanism_type = line[titles.index("KineticMechanismType")]
        parameter_unit = line[titles.index("parameter.unit")]
        parameter_associated_species = line[titles.index("parameter.associatedSpecies")]
        substrates = line[titles.index("Substrate")].split(";")
        products = line[titles.index("Product")].split(";")
        chebi_ids = line[titles.index("ChebiID")].split(";")
        try:
            temperature = float(line[titles.index("Temperature")])
        except (ValueError, IndexError):
            temperature = None
        try:
            ph = float(line[titles.index("pH")])
        except (ValueError, IndexError):
            ph = None

        if ec_number not in sabio_dict_pointer:
            sabio_dict_pointer[ec_number] = []
        sabio_dict_pointer[ec_number].append(
            SabioEntry(
                entry_id=entry_id,
                is_recombinant=is_recombinant,
                kinetics_mechanism_type=kinetics_mechanism_type,
                organism=organism,
                temperature=temperature,
                ph=ph,
                parameter_unit=parameter_unit,
                parameter_value=parameter_value,
                parameter_associated_species=parameter_associated_species,
                substrates=substrates,
                products=products,
                chebi_ids=chebi_ids,
            )
        )
    return sabio_dict


def sabio_select_enzyme_kinetic_data_for_model(
    cobra_model: cobra.Model,
    sabio_target_folder: str,
    base_species: str,
    ncbi_parsed_json_path: str,
    bigg_metabolites_json_path: str,
    kinetic_ignored_metabolites: list[str] = [],
    kinetic_ignored_enzyme_ids: list[str] = [],
    custom_enzyme_kinetic_data: dict[str, EnzymeReactionData | None] = {},
    min_ph: float = -float("inf"),
    max_ph: float = float("inf"),
    accept_nan_ph: bool = True,
    min_temperature: float = -float("inf"),
    max_temperature: float = float("inf"),
    accept_nan_temperature: bool = True,
    kcat_overwrite: dict[str, float] = {},
    transfered_ec_number_json: str = "",
    max_taxonomy_level: int = float("inf"),
) -> dict[str, EnzymeReactionData | None]:
    """Selects enzyme kinetic data for a given COBRA model using SABIO-RK data.

    If this data cannot be found, an internet connection is built to SABIO-RK and the relevant
    data is downloaded, which may take some time in the order of dozens of minutes.
    If you want to download the full SABIO-RK data beforehand, run get_full_sabio_dict() from
    this module beforehand, with the same sabio_target_folder.

    Collected data includes k_cat, k_m, k_i, k_a and Hill coefficients for all EC numbers that
    occur in the model's BiGG-compliant EC number annotation.

    Args:
        cobra_model (cobra.Model): The COBRA model for which enzyme kinetic data is to be selected.
        sabio_target_folder (str): The path to the folder containing SABIO-RK data.
        base_species (str): The base species for taxonomy comparison.
        ncbi_parsed_json_path (str): The path to the NCBI parsed JSON file.
        bigg_metabolites_json_path (str): The path to the BIGG metabolites JSON file.
        kinetic_ignored_metabolites (list[str], optional): List of metabolites to ignore. Defaults to [].
        kinetic_ignored_enzyme_ids (list[str], optional): List of enzyme IDs to ignore. Defaults to [].
        custom_enzyme_kinetic_data (dict[str, EnzymeReactionData | None], optional): Custom enzyme kinetic data. Defaults to {}.
        min_ph (float, optional): Minimum pH value for filtering. Defaults to -float("inf").
        max_ph (float, optional): Maximum pH value for filtering. Defaults to float("inf").
        accept_nan_ph (bool, optional): Whether to accept entries with NaN pH values. Defaults to True.
        min_temperature (float, optional): Minimum temperature value for filtering. Defaults to -float("inf").
        max_temperature (float, optional): Maximum temperature value for filtering. Defaults to float("inf").
        accept_nan_temperature (bool, optional): Whether to accept entries with NaN temperature values. Defaults to True.
        kcat_overwrite (dict[str, float], optional): Dictionary to overwrite kcat values. Defaults to {}.

    Returns:
        dict[str, EnzymeReactionData | None]: A dictionary mapping reaction IDs to enzyme kinetic data.

    Example:
        cobra_model = cobra.io.read_sbml_model("model.xml")
        sabio_target_folder = "/path/to/sabio/data"
        base_species = "Escherichia coli"
        ncbi_parsed_json_path = "/path/to/ncbi.json"
        bigg_metabolites_json_path = "/path/to/bigg_metabolites.json"
        enzyme_kinetic_data = sabio_select_enzyme_kinetic_data_for_model(
            cobra_model,
            sabio_target_folder,
            base_species,
            ncbi_parsed_json_path,
            bigg_metabolites_json_path
        )
        print(enzyme_kinetic_data)
    """
    sabio_dict = get_full_sabio_dict(
        sabio_target_folder,
    )
    ncbi_parsed_json_data = json_zip_load(ncbi_parsed_json_path)
    name_to_bigg_id_dict: dict[str, str] = json_load(
        bigg_metabolites_json_path, dict[str, str]
    )

    # Get reaction<->enzyme reaction data mapping
    enzyme_reaction_data: dict[str, EnzymeReactionData | None] = {}
    transfered_ec_codes: dict[str, str] = (
        json_load(transfered_ec_number_json, dict[str, str])
        if transfered_ec_number_json
        else {}
    )
    for reaction in cobra_model.reactions:
        if "ec-code" not in reaction.annotation:
            continue

        enzyme_identifiers = reaction.gene_reaction_rule.split(" and ")
        has_found_ignored_enzyme = False
        for enzyme_identifier in enzyme_identifiers:
            if enzyme_identifier in kinetic_ignored_enzyme_ids:
                has_found_ignored_enzyme = True
                break
        if has_found_ignored_enzyme:
            continue

        reac_met_ids = [met.id for met in reaction.metabolites]
        substrate_bigg_ids = [
            met_id[: met_id.rfind("_")]
            for met_id in reac_met_ids
            if reaction.metabolites[cobra_model.metabolites.get_by_id(met_id)] < 0
        ]
        product_bigg_ids = [
            met_id[: met_id.rfind("_")]
            for met_id in reac_met_ids
            if reaction.metabolites[cobra_model.metabolites.get_by_id(met_id)] > 0
        ]

        ec_codes = reaction.annotation["ec-code"]
        if isinstance(ec_codes, str):
            ec_codes = [ec_codes]
        reaction_transfered_ec_codes = [
            transfered_ec_codes[ec_code]
            for ec_code in ec_codes
            if ec_code in transfered_ec_codes
        ]
        ec_codes += reaction_transfered_ec_codes

        all_entries = (
            (
                "kcat",
                _get_ec_code_entries(
                    sabio_dict.kcat_entries,
                    ec_codes,
                    min_ph,
                    max_ph,
                    accept_nan_ph,
                    min_temperature,
                    max_temperature,
                    accept_nan_temperature,
                    substrate_bigg_ids,
                    product_bigg_ids,
                    name_to_bigg_id_dict,
                ),
            ),
            (
                "km",
                _get_ec_code_entries(
                    sabio_dict.km_entries,
                    ec_codes,
                    min_ph,
                    max_ph,
                    accept_nan_ph,
                    min_temperature,
                    max_temperature,
                    accept_nan_temperature,
                    substrate_bigg_ids,
                    product_bigg_ids,
                    name_to_bigg_id_dict,
                ),
            ),
            (
                "ki",
                _get_ec_code_entries(
                    sabio_dict.ki_entires,
                    ec_codes,
                    min_ph,
                    max_ph,
                    accept_nan_ph,
                    min_temperature,
                    max_temperature,
                    accept_nan_temperature,
                    substrate_bigg_ids,
                    product_bigg_ids,
                    name_to_bigg_id_dict,
                ),
            ),
            (
                "ka",
                _get_ec_code_entries(
                    sabio_dict.ka_entires,
                    ec_codes,
                    min_ph,
                    max_ph,
                    accept_nan_ph,
                    min_temperature,
                    max_temperature,
                    accept_nan_temperature,
                    substrate_bigg_ids,
                    product_bigg_ids,
                    name_to_bigg_id_dict,
                ),
            ),
        )

        # {'mol', 'katal*g^(-1)', 'M', 'M^2', 'g', 'mol/mol', 'J/mol', '-',
        # 's^(-1)', 's^(-1)*g^(-1)', 'mg/ml', 'mol*s^(-1)*mol^(-1)', 'M^(-1)', 'Pa',
        # 'M^(-1)*s^(-1)', 'mol*s^(-1)*g^(-1)', 'katal'}
        k_cat_per_tax_score: dict[int, list[float]] = {}
        k_cat_refs_per_tax_score: dict[int, list[ParameterReference]] = {}
        k_ms_per_tax_score: dict[str, dict[int, list[float]]] = {}
        k_m_refs_per_tax_score: dict[str, dict[int, list[ParameterReference]]] = {}
        k_is_per_tax_score: dict[str, dict[int, list[float]]] = {}
        k_i_refs_per_tax_score: dict[str, dict[int, list[ParameterReference]]] = {}
        k_as_per_tax_score: dict[str, dict[int, list[float]]] = {}
        k_a_refs_per_tax_score: dict[str, dict[int, list[ParameterReference]]] = {}
        for entries_type, entries in all_entries:
            if entries_type == "kcat":  # Reaction-wide search
                for entry in entries:
                    match entry.parameter_unit:
                        case "s^(-1)":
                            multiplier = 3_600
                        case _:
                            continue

                    taxonomy_dict = get_taxonomy_dict_from_nbci_taxonomy(
                        [base_species, entry.organism], ncbi_parsed_json_data
                    )
                    taxonomy_score = get_taxonomy_scores(base_species, taxonomy_dict)[
                        entry.organism
                    ]
                    if taxonomy_score > max_taxonomy_level:
                        continue
                    if taxonomy_score not in k_cat_per_tax_score:
                        k_cat_per_tax_score[taxonomy_score] = []
                        k_cat_refs_per_tax_score[taxonomy_score] = []
                    k_cat_per_tax_score[taxonomy_score].append(
                        entry.parameter_value * multiplier
                    )
                    k_cat_refs_per_tax_score[taxonomy_score].append(
                        ParameterReference(
                            database="SABIO-RK",
                            comment="SabioEntryID: " + str(entry.entry_id),
                            species=entry.organism,
                            substrate=entry.parameter_associated_species,
                            value=entry.parameter_value * multiplier,
                        )
                    )
            else:  # Metabolite-wide search
                match entries_type:
                    case "ka":
                        values_pointer = k_as_per_tax_score
                        ref_pointer = k_a_refs_per_tax_score
                    case "ki":
                        values_pointer = k_is_per_tax_score
                        ref_pointer = k_i_refs_per_tax_score
                    case "km":
                        values_pointer = k_ms_per_tax_score
                        ref_pointer = k_m_refs_per_tax_score
                    case _:
                        raise ValueError
                for met in cobra_model.metabolites:
                    if met.id in kinetic_ignored_metabolites:
                        continue
                    if (entries_type == "km") and met not in reaction.metabolites:
                        continue
                    bigg_id = met.id[: met.id.rfind("_")]
                    for entry in entries:
                        entry_met_id = (
                            entry.parameter_associated_species.lower().strip()
                        )
                        if entry_met_id in name_to_bigg_id_dict:
                            entry_bigg_id = name_to_bigg_id_dict[entry_met_id]
                        else:
                            entry_bigg_id = _search_metname_in_bigg_ids(
                                met_id=entry_met_id,
                                bigg_id="",
                                entry=entry,
                                name_to_bigg_id_dict=name_to_bigg_id_dict,
                            )
                            if not entry_bigg_id:
                                continue
                        if entry_bigg_id != bigg_id:
                            continue

                        match entry.parameter_unit:
                            case "M^2":
                                applier = sqrt
                            case "M^(-1)":
                                applier = lambda x: 1 / x  # noqa: E731
                            case "M":
                                applier = lambda x: x  # noqa: E731
                            case _:
                                continue
                        taxonomy_dict = get_taxonomy_dict_from_nbci_taxonomy(
                            [base_species, entry.organism], ncbi_parsed_json_data
                        )
                        taxonomy_score = get_taxonomy_scores(
                            base_species, taxonomy_dict
                        )[entry.organism]
                        if taxonomy_score > max_taxonomy_level:
                            continue

                        if met.id not in values_pointer:
                            values_pointer[met.id] = {}
                            ref_pointer[met.id] = {}
                        if taxonomy_score not in values_pointer[met.id]:
                            values_pointer[met.id][taxonomy_score] = []
                            ref_pointer[met.id][taxonomy_score] = []
                        values_pointer[met.id][taxonomy_score].append(
                            applier(entry.parameter_value)
                        )
                        ref_pointer[met.id][taxonomy_score].append(
                            ParameterReference(
                                database="SABIO-RK",
                                comment="SabioEntryID: " + str(entry.entry_id),
                                species=entry.organism,
                                substrate=entry.parameter_associated_species,
                                tax_distance=taxonomy_score,
                                value=applier(entry.parameter_value),
                            )
                        )

        if reaction.id in kcat_overwrite:
            k_cat = kcat_overwrite[reaction.id]
            k_cat_references = [
                ParameterReference(database="OVERWRITE", tax_distance=-1)
            ]
        elif (
            (reaction.id not in kcat_overwrite) and (kcat_overwrite != {})
        ) or not k_cat_per_tax_score:
            continue
        else:
            min_k_cat_tax_score = min(k_cat_per_tax_score.keys())
            k_cat = median(k_cat_per_tax_score[min_k_cat_tax_score])
            k_cat_references = k_cat_refs_per_tax_score[min_k_cat_tax_score]

        k_ms: dict[str, float] = {}
        k_m_references: dict[str, list[ParameterReference]] = {}
        for met_id, k_m_per_tax_score in k_ms_per_tax_score.items():
            k_ms[met_id] = median(k_m_per_tax_score[min(k_m_per_tax_score.keys())])
            k_m_references[met_id] = k_m_refs_per_tax_score[met_id][
                min(k_m_per_tax_score.keys())
            ]

        k_is: dict[str, float] = {}
        k_i_references: dict[str, list[ParameterReference]] = {}
        for met_id, k_i_per_tax_score in k_is_per_tax_score.items():
            k_is[met_id] = median(k_i_per_tax_score[min(k_i_per_tax_score.keys())])
            k_i_references[met_id] = k_i_refs_per_tax_score[met_id][
                min(k_i_per_tax_score.keys())
            ]

        k_as: dict[str, float] = {}
        k_a_references: dict[str, list[ParameterReference]] = {}
        for met_id, k_a_per_tax_score in k_as_per_tax_score.items():
            k_as[met_id] = median(k_a_per_tax_score[min(k_a_per_tax_score.keys())])
            k_a_references[met_id] = k_a_refs_per_tax_score[met_id][
                min(k_a_per_tax_score.keys())
            ]

        enzyme_reaction_data[reaction.id] = EnzymeReactionData(
            identifiers=enzyme_identifiers,
            k_cat=k_cat,
            k_cat_references=k_cat_references,
            k_ms=k_ms,
            k_m_references=k_m_references,
            k_is=k_is,
            k_i_references=k_i_references,
            k_as=k_as,
            k_a_references=k_a_references,
        )

    enzyme_reaction_data = {**enzyme_reaction_data, **custom_enzyme_kinetic_data}

    for reac_id in kcat_overwrite:  # noqa: PLC0206
        if reac_id not in enzyme_reaction_data:
            reaction = cobra_model.reactions.get_by_id(reac_id)
            enzyme_identifiers = reaction.gene_reaction_rule.split(" and ")
            enzyme_reaction_data[reac_id] = EnzymeReactionData(
                identifiers=enzyme_identifiers,
                k_cat=kcat_overwrite[reac_id],
                k_cat_references=[
                    ParameterReference(database="OVERWRITE", tax_distance=-1)
                ],
                k_ms={},
                k_is={},
            )

    return enzyme_reaction_data



"""Currently just contains the toy model form COBRAk's documentation as example model"""

# IMPORT SECTION
from math import log

from .constants import STANDARD_R, STANDARD_T
from .dataclasses import (
    Enzyme,
    EnzymeReactionData,
    ExtraLinearConstraint,
    Metabolite,
    Model,
    Reaction,
)

# EXAMPLE MODEL DEFINITION SECTION
toy_model = Model(
    reactions={
        # Metabolic reactions
        "Glycolysis": Reaction(
            # Stoichiometrically relevant member variables
            stoichiometries={
                "S": -1,  # Negative stoichiometry → S is consumed by Glycolysis
                "M": +1,  # Positive stoichiometry → M is produced by Glycolysis
                "ATP": +2,  # Two ATP molecules are produced by Glycolysis
            },
            min_flux=0.0,  # Minimal flux in mmol⋅gDW⁻¹⋅h⁻¹; should be ≥0 for most analyses
            max_flux=1_000.0,  # Maximal flux in mmol⋅gDW⁻¹⋅h⁻¹
            # Thermodynamically relevant member variables
            # (only neccessary if thermodynamic constraints are used)
            dG0=-10.0,  # Standard Gibb's free energy ΔG'° in kJ⋅mol⁻¹; Default is None (no ΔG'°)
            dG0_uncertainty=None,  # ΔG'° uncertainty in kJ⋅mol⁻¹; Default is None (no uncertainty)
            # Let's set the variable for enzyme-kinetic parameters
            # of the dataclass EnzymeReactionData
            # (Default is None, i.e. no enzyme parameters given)
            enzyme_reaction_data=EnzymeReactionData(
                identifiers=[
                    "E_glyc"
                ],  # Subunit(s) which constitute the reaction's catalyst
                k_cat=70_000.0,  # Turnover number in h⁻¹
                k_ms={  # Michaelis-Menten constants in M=mol⋅l⁻¹; Default is {}
                    "S": 0.0001,  # e.g., K_m of reaction Glycolysis regarding metabolite A
                    "M": 0.0001,
                    "ATP": 0.0001,
                },
                special_stoichiometries={},  # No special stoichiometry, all subunits occur once
            ),
            # Extra information member variables
            annotation={"description": "This is reaction Glycolysis"},  # Default is {}
            name="Reaction Glycolysis",  # Default is ""
        ),
        "Respiration": Reaction(
            stoichiometries={
                "M": -1,
                "C": +1,
                "ATP": +6,
            },
            min_flux=0.0,
            max_flux=1_000.0,
            dG0=-10.0,
            enzyme_reaction_data=EnzymeReactionData(
                identifiers=["E_resp"],
                k_cat=70_000.0,
                k_ms={
                    "M": 0.03,
                    "C": 0.0001,
                    "ATP": 0.0001,
                },
            ),
        ),
        "Overflow": Reaction(
            stoichiometries={
                "M": -1,
                "P": +1,
            },
            min_flux=0.0,
            max_flux=1_000.0,
            dG0=-10.0,
            enzyme_reaction_data=EnzymeReactionData(
                identifiers=["E_over"],
                k_cat=70_000.0,
                k_ms={
                    "M": 0.005,
                    "P": 0.0001,
                },
            ),
        ),
        # Exchange reactions
        "EX_S": Reaction(
            stoichiometries={
                "S": +1,
            },
            min_flux=0.0,
            max_flux=1_000.0,
        ),
        "EX_C": Reaction(
            stoichiometries={
                "C": -1.0,
            },
            min_flux=0.0,
            max_flux=1_000.0,
        ),
        "EX_P": Reaction(
            stoichiometries={
                "P": -1,
            },
            min_flux=0.0,
            max_flux=1_000.0,
        ),
        "EX_ATP": Reaction(
            stoichiometries={
                "ATP": -1,
            },
            min_flux=0.0,
            max_flux=1_000.0,
        ),
    },
    metabolites={
        "S": Metabolite(
            log_min_conc=log(
                1e-6
            ),  # optional, minimal ln(concentration); Default is ln(1e-6 M)
            log_max_conc=log(
                0.02
            ),  # optional, maximal ln(concentration); Default is ln(0.02 M)
            annotation={
                "description": "This is metabolite S"
            },  # optional, default is ""
            name="Metabolite S",  # optional, default is ""
            formula="X",  # optional, default is ""
            charge=0,  # optional, default is 0
        ),
        "M": Metabolite(),
        "C": Metabolite(),
        "P": Metabolite(),
        "ATP": Metabolite(),
    },
    enzymes={
        "E_glyc": Enzyme(
            molecular_weight=1_000.0,  # Molecular weight in kDa
            min_conc=None,  # Optional concentration in mmol⋅gDW⁻¹; Default is None (minimum is 0)
            max_conc=None,  # Optional maximal concentration in mmol⋅gDW⁻¹; Default is None (only protein pool restricts)
            annotation={"description": "Enzyme of Glycolysis"},  # Default is {}
            name="Glycolysis enzyme",  # Default is ""
        ),
        "E_resp": Enzyme(molecular_weight=2_500.0),
        "E_over": Enzyme(molecular_weight=500.0),
    },
    max_prot_pool=0.25,  # In g⋅gDW⁻¹; This value is used for our analyses with enzyme constraints
    # We set the following two constraints:
    # 1.0 * EX_A - 1.0 * Glycolysis ≤ 0.0
    # and
    # 1.0 * EX_A + 1.0 * Glycolysis ≥ 0.0
    # in other words, effectively,
    # 1.0 * EX_A = 1.0 * Glycolysis
    extra_linear_constraints=[
        ExtraLinearConstraint(
            stoichiometries={
                "EX_S": -1.0,
                "Glycolysis": 1.0,
            },
            lower_value=0.0,
            upper_value=0.0,
        )
    ],  # Keep in mind that this is a list as multiple extra flux constraints are possible
    kinetic_ignored_metabolites=[],
    R=STANDARD_R,
    T=STANDARD_T,
    max_conc_sum=float("inf"),
    annotation={"description": "COBRAk toy model"},
)



from copy import deepcopy
from random import choice

from joblib import Parallel, cpu_count, delayed
from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    Objective,
    Reals,
    SolverFactoryClass,
    Var,
    maximize,
)

from .constants import BIG_M, STANDARD_MIN_MDF
from .dataclasses import Model, Solver
from .lps import (
    get_lp_from_cobrak_model,
    get_objective,
)
from .pyomo_functionality import get_solver
from .standard_solvers import SCIP
from .utilities import (
    delete_orphaned_metabolites_and_enzymes,
    delete_unused_reactions_in_variability_dict,
    get_stoichiometrically_coupled_reactions,
)

STANDARD_EFM_OBJ_NAME = "EFM_STANDARD_OBJECTIVE"


def _efm_batch_routine(
    pyomo_model: ConcreteModel,
    pyomo_solver: SolverFactoryClass,
    objective_name: str,
    cobrak_solver: Solver,
    objective_sense: int,
) -> None:
    getattr(pyomo_model, objective_name).activate()

    # OBJ ACTIVATE
    pyomo_solver.solve(
        pyomo_model, tee=True, warmstart=True, **cobrak_solver.solve_extra_options
    )

    if results.solver.termination_condition == TerminationCondition.maxTimeLimit:
        result = None
    else:
        result = value(getattr(model, target_id))

    getattr(pyomo_model, objective_name).deactivate()


def calculate_efms(
    cobrak_model: Model,
    with_inhomogenous_constraints: bool = False,
    with_enzyme_constraints: bool = False,
    with_thermodynamic_constraints: bool = False,
    solver: Solver = SCIP,
    variability_dict: dict[str, tuple[float, float]] = {},
    max_num_efms: int = 1_000_000_000,
    min_mdf: float = STANDARD_MIN_MDF,
    min_efm_flux_bound: float = 1e-5,
) -> set[tuple[str, ...]]:
    enforced_reacs = [
        reac_id
        for reac_id in variability_dict
        if variability_dict[reac_id][0] > 0.0 and reac_id in cobrak_model.reactions
    ]
    if not with_inhomogenous_constraints:
        for enforced_reac in enforced_reacs:
            cobrak_model.reactions[enforced_reac].min_flux = 0.0
            variability_dict[enforced_reac] = (0.0, variability_dict[enforced_reac][1])

    reduced_model = deepcopy(cobrak_model)
    if variability_dict != {}:
        reduced_model = delete_unused_reactions_in_variability_dict(
            reduced_model, variability_dict
        )
    reduced_model = delete_orphaned_metabolites_and_enzymes(reduced_model)

    all_couples = get_stoichiometrically_coupled_reactions(
        reduced_model,
    )
    eligible_couples: list[list[str]] = []
    for couple in all_couples:
        eligible = True
        for enforced_reac in enforced_reacs:
            if enforced_reac in couple:
                eligible = False
                break
        if eligible:
            eligible_couples.append(couple)

    pyomo_model = get_lp_from_cobrak_model(
        cobrak_model=reduced_model,
        with_enzyme_constraints=with_enzyme_constraints,
        with_thermodynamic_constraints=with_thermodynamic_constraints,
        with_loop_constraints=True,
        min_mdf=min_mdf,
    )
    zc_var_sum = 0.0
    zc_var_ids: list[str] = []
    objective_names: list[str] = []
    for eligible_couple in eligible_couples:
        zc_var_id = "zC_var_" + "_".join(eligible_couple)
        zc_var_ids.append(zc_var_id)

        setattr(pyomo_model, zc_var_id, Var(within=Binary))
        setattr(
            pyomo_model,
            zc_var_id + "_constraint",
            Constraint(
                rule=getattr(pyomo_model, eligible_couple[0])
                <= BIG_M * getattr(pyomo_model, zc_var_id)
            ),
        )
        setattr(
            pyomo_model,
            zc_var_id + "_constraint_2",
            Constraint(
                rule=getattr(pyomo_model, eligible_couple[0])
                >= min_efm_flux_bound * getattr(pyomo_model, zc_var_id)
            ),
        )
        zc_var_sum += getattr(pyomo_model, zc_var_id)

        objective_names.append(f"objective_{eligible_couple[0]}")
        setattr(
            pyomo_model,
            objective_names[-1],
            Objective(expr=getattr(pyomo_model, eligible_couple[0]), sense=maximize),
        )
        getattr(pyomo_model, objective_names[-1]).deactivate()

    objective_names.append(f"objective_{eligible_couple[0]}")
    setattr(
        pyomo_model,
        STANDARD_EFM_OBJ_NAME,
        Objective(expr=0.0, sense=maximize),
    )
    getattr(pyomo_model, STANDARD_EFM_OBJ_NAME).deactivate()

    setattr(pyomo_model, "zc_var_sum", Var(within=Reals, bounds=(1.0, 10_000.0)))
    setattr(
        pyomo_model,
        "zc_var_sum_constraint_1",
        Constraint(rule=getattr(pyomo_model, "zc_var_sum") == zc_var_sum),
    )
    pyomo_model.obj = get_objective(pyomo_model, "zc_var_sum", -1)

    integer_cut_i = 0
    pyomo_solver = get_solver(solver.name, solver.solver_options, solver.solver_attrs)
    efms: set[tuple[str, ...]] = set()
    for i in range(max_num_efms):
        opt_targets = [STANDARD_EFM_OBJ_NAME] + [
            choice(objective_names) for _ in range(cpu_count() - 1)
        ]
        min_z_results = Parallel(n_jobs=-1, verbose=10)(
            delayed(_efm_batch_routine)(
                pyomo_model,
                pyomo_solver,
                opt_target,
                +1,
            )
            for opt_target in opt_targets
        )

        new_efms = min_z_results.difference(efms)
        if len(new_efms) == 0:
            break
        for new_efm in new_efms:
            new_integer_cut = 0.0
            new_integer_cut_sum = 0.0
            for zc_var_id in zc_var_ids:
                if getattr(pyomo_model, zc_var_id).value <= 0.1:
                    continue
                new_integer_cut += getattr(pyomo_model, zc_var_id)
                new_integer_cut_sum += 1
            setattr(
                pyomo_model,
                f"integer_cut_{integer_cut_i}",
                Constraint(expr=new_integer_cut <= new_integer_cut_sum - 1),
            )
            integer_cut_i += 1
        efms = efms.union(new_efms)

    return efms



"""This module contains all COBRAk constants that are used throughout its packages.

These constants are especially used for problem constructions (to determine prefixes,
suffixes, names, ... for pyomo variables) as well as thermodynamic standard values.
"""

# CONSTANTS #
ALL_OK_KEY = "ALL_OK"
"""Shows that the result is optimal and the termination condition is ok"""

BIG_M = 10_000
"""Big M value for MILPs"""

DF_VAR_PREFIX = "f_var_"
"""Prefix for driving force problem variables"""

DG0_VAR_PREFIX = "dG0_"
"""Prefix for Gibb's free energy problem variables"""


EC_INNER_TO_OUTER_COMPARTMENTS: list[str] = [
    "c",
    "p",
    "e",
]
"""Inner to outer compartments in this order for E. coli models used here"""


EC_IONIC_STRENGTHS: dict[str, float] = {
    "c": 250,  # Source: eQuilibrator standard
    "p": 250,
    "e": 250,
}
"""Ionic strenghts (in mM) for E. coli model compartments used here"""


EC_PHS: dict[str, float] = {
    "c": 7.5,  # Source: Bionumbers ID 105980
    "p": 7.5,  # Source: Bionumbers ID 105980
    "e": 7.5,
}
"""pH values (unitless) for E. coli model compartments used here"""

EC_PMGS: dict[str, float] = {
    "c": 2.5,  # Source: eQuilibrator standard
    "p": 2.5,
    "e": 2.5,
}
"""pMg values (unitless) for E. coli model compartments used here"""

EC_POTENTIAL_DIFFERENCES: dict[tuple[str, str], float] = {
    ("c", "p"): 0.15,  # Source: eQuilibrator standard
    ("p", "e"): 0.15,
}
"""Potential differences (in V) for E. coli model compartments used here"""

ENZYME_VAR_PREFIX = "enzyme_"
"""Prefix of problem variables which stand for enzyme concentrations"""

ENZYME_VAR_INFIX = "_of_"
"""Infix for separation of enzyme name and reaction name"""

ERROR_CONSTRAINT_PREFIX = "flux_error_"
"""Prefix for the constraint that defines a scenario flux constraint"""

ERROR_BOUND_LOWER_CHANGE_PREFIX = "bound_error_change_lower_"
"""Prefix for fixed variables that show how much affected lower variable bounds have to be changed"""

ERROR_BOUND_UPPER_CHANGE_PREFIX = "bound_error_change_upper_"
"""Prefix for fixed variables that show how much affected lower variable bounds have to be changed"""

ERROR_VAR_PREFIX = "error_"
"""Prefix for error term variables for feasibility-making optimizations"""

ERROR_SUM_VAR_ID = "error_sum"
"""Name for the variable that holds the sum of all error term variables"""

FLUX_SUM_VAR_ID = "FLUX_SUM_VAR"
"""Name of optional variable that holds the sum of all reaction fluxes"""

KAPPA_PRODUCTS_VAR_PREFIX = "kappa_products_"
"""Prefix for variables representing the sum of logairthmized product concentration minus the logarithmized sum of km values"""

KAPPA_SUBSTRATES_VAR_PREFIX = "kappa_substrates_"
"""Prefix for variables representing the sum of logairthmized substrate concentration minus the logarithmized sum of km values"""

KAPPA_VAR_PREFIX = "kappa_var_"
"""Prefix for variables representing the thermodynamic restriction of a reaction (used in non-linear programs)"""

GAMMA_VAR_PREFIX = "gamma_var_"
"""Prefix for variables representing the thermodynamic restriction of a reaction (used in non-linear programs)"""

IOTA_VAR_PREFIX = "iota_var_"
"""Prefix for variables representing the inhibition of a reaction (used in non-linear programs)"""

ALPHA_VAR_PREFIX = "alpha_var_"
"""Prefix for variables representing the activation of a reaction (used in non-linear programs)"""

LNCONC_VAR_PREFIX = "x_"
"""Prefix for logarithmized concentration problem variables"""

MDF_VAR_ID = "var_B"
"""Name for minimally occuring driving force variable"""

OBJECTIVE_CONSTRAINT_NAME = "objective_constraint"
"""Name for constraint that defines the objective function's term"""

OBJECTIVE_VAR_NAME = "OBJECTIVE_VAR"
"""Name for variable that holds the objective value"""

PROT_POOL_MET_NAME = "prot_pool"
"""Identifier of the protein pool representing pseudo-metabolite"""

PROT_POOL_REAC_NAME = PROT_POOL_MET_NAME + "_delivery"
"""Identifier of the pseudo-reaction which created the protein pool pseudo-metabolite"""

QUASI_INF = 100_000
"""Big number (larger than big M) for values that would reach inf (thereby potentially causing solver problems)"""

REAC_ENZ_SEPARATOR = "_ENZ_"
"""Separator between enzyme-constrained reaction ID and attached enzyme name"""

REAC_FWD_SUFFIX = "_FWD"
"""Standard suffix for reaction IDs that represent forward directions of originally irreversible reactions"""

REAC_REV_SUFFIX = "_REV"
"""Standard suffix for reaction IDs that represent reverse directions of originally irreversible reactions"""

SOLVER_STATUS_KEY = "SOLVER_STATUS"
"""Solver status optimization dict key"""

STANDARD_MIN_MDF = 1e-3
"""Standard minimally ocurring driving force for active reactions in kJ⋅mol⁻¹"""

STANDARD_R = 8.314e-3
"""Standard gas constant in kJ⋅K⁻1⋅mol⁻1 (Attention: Standard value is often given in J⋅K⁻1⋅mol⁻1, but we need in kJ⋅K⁻1⋅mol⁻1)"""

STANDARD_T = 298.15
"""Standard temperature in Kelvin"""

TERMINATION_CONDITION_KEY = "TERMINATION_CONDITION"
"""Solver termination condition key in optimization dict"""

USED_IDENTIFIERS_FOR_EQUILIBRATOR = [
    "inchi",
    "inchi_key",
    "metanetx.chemical",
    "bigg.metabolite",
    "kegg.compound",
    "chebi",
    "sabiork.compound",
    "metacyc.compound",
    "hmdb",
    "swisslipid",
    "reactome",
    "lipidmaps",
    "seed.compound",
]
"""Standard bunch of reaction identifier annotation names for E. coli models used here"""

Z_VAR_PREFIX = "z_var_"
"""Prefix of z variables (used with thermodynamic constraints in MI(N)LPs)"""



from platform import platform

from joblib import cpu_count

from .dataclasses import Solver

BARON = Solver(
    name="gams",
    solver_options={
        "solver": "baron",
    },
    solve_extra_options={
        "add_options": [
            "GAMS_MODEL.optfile = 1;",
            "$onecho > baron.opt",
            "$offecho",
        ]
    },
)

CPLEX = Solver(
    name="cplex_direct",
)

CPLEX_FOR_VARIABILITY_ANALYSIS = Solver(
    name="cplex_direct",
    solver_options={
        "threads": 1,
        "lpmethod": 1,
    },
)

GUROBI = Solver(
    name="gurobi_direct",
)

GUROBI_FOR_VARIABILITY_ANALYSIS = Solver(
    name="gurobi_direct",
    solver_options={
        "threads": 1,
    },
)

hsllib = "/scratch/CoinHSL/coinhsl-2024.05.15/BUILD/lib/x86_64-linux-gnu/libcoinhsl.so"
if platform().startswith("macOS"):
    hsllib = "/Users/pbekiaris/0Programs/CoinHSL.v2024.5.15.aarch64-apple-darwin-libgfortran5/lib/libcoinhsl.dylib"
elif cpu_count() > 16:
    hsllib = "/u/pbekiaris/CoinHSL_2024_BUILT/lib/x86_64-linux-gnu/libcoinhsl.so"
elif cpu_count() > 6:
    hsllib = (
        "/mechthild/home/bekiaris/CoinHSL_2024_BUILT/lib/x86_64-linux-gnu/libcoinhsl.so"
    )

IPOPT_MA57 = Solver(
    name="ipopt",
    solver_options={
        "max_iter": 4_000,
        "halt_on_ampl_error": "yes",
        "mu_strategy": "adaptive",
        "corrector_type": "primal-dual",
        "acceptable_tol": 1e-8,  # same as default "tol"
        "acceptable_constr_viol_tol": 0.0001,  # same as default "constr_viol_tol"
        "acceptable_dual_inf_tol": 1.0,  # same as default "dual_inf_tol"
        "acceptable_iter": 0,
        "linear_solver": "ma57",
        "hsllib": hsllib,
        "nlp_scaling_method": "none",
        "ma57_automatic_scaling": "no",
        "ma57_block_size": "32",
        "ma97_small": 1e-30,
        "ma97_u": 1e-6,
    },
)

IPOPT = Solver(
    name="ipopt",
    solver_options={
        "max_iter": 4_000,
        "halt_on_ampl_error": "yes",
    },
)

IPOPT_LONGRUN = Solver(
    name="ipopt",
    solver_options={
        "max_iter": 100_000,
        "halt_on_ampl_error": "yes",
        "mu_strategy": "adaptive",
        "corrector_type": "primal-dual",
        "linear_solver": "ma57",
        "hsllib": hsllib,
        "acceptable_tol": 1e-8,  # same as default "tol"
        "acceptable_constr_viol_tol": 0.0001,  # same as default "constr_viol_tol"
        "acceptable_dual_inf_tol": 1.0,  # same as default "dual_inf_tol"
        "acceptable_iter": 0,
        "nlp_scaling_method": "none",
        "ma57_automatic_scaling": "no",
    },
)

SCIP = Solver(
    name="scip",
    solver_attrs={"_version_timeout": 180},
)



"""This script is a wrapper for the ΔG'° determination with the eQuilibrator API.

This wrapper intends to work with BiGG-styled cobrapy metabolic models.
"""

# IMPORTS SECTION #
from typing import Any

import cobra
from equilibrator_api import Q_, ComponentContribution, Reaction

from .constants import USED_IDENTIFIERS_FOR_EQUILIBRATOR


## PUBLIC FUNCTIONS ##
def equilibrator_get_model_dG0_and_uncertainty_values(
    sbml_path: str,
    inner_to_outer_compartments: list[str],
    phs: dict[str, float],
    pmgs: dict[str, float],
    ionic_strengths: dict[str, float],
    potential_differences: dict[tuple[str, str], float],
    exclusion_prefixes: list[str] = [],
    exclusion_inner_parts: list[str] = [],
    ignore_uncertainty: bool = False,
    max_uncertainty: float = 1_000.0,
) -> tuple[dict[str, float], dict[str, float]]:
    """Cobrapy model wrapper for the ΔG'° determination of reactions using the eQuilibrator-API.

    Reactions are identified according to all annotation (in the cobrapy reaction's annotation member variable)
    given in this modules global USED_IDENTIFIERS list.

    Args:
        sbml_path (str): The path to the SBML-encoded constraint-based metabolic model for which ΔG'° values are determined.
        inner_to_outer_compartments (List[str]): A list with compartment IDs going from inner (e.g., in E. coli,
            the cytosol or 'c' in iML1515) to outer (e.g., the extracellular component or 'e' in iML1515). Used
            for the ΔG'° calculation in multi-compartmental reactions.
        phs (Dict[str, float]): A dictionary with compartment IDs as keys and the compartment pHs as values.
        pmgs (Dict[str, float]): A dictionary with compartment IDs as keys and the compartment pMgs as values.
        ionic_strengths (Dict[str, float]): A dictionary with compartment IDs as keys and the ionic strengths as values.
        potential_differences (Dict[Tuple[str, str], float]): A dictionary containing tuples with 2 elements describing
            the ID of an innter and outer compartment, and the potential difference between them.
        max_uncertainty (float): The maximal accepted uncertainty value (defaults to 1000 kJ⋅mol⁻¹). If a calculated uncertainty
            is higher than this value, the associated ΔG'° is *not* used (i.e., the specific reaction gets no ΔG'°).

    Returns:
        Dict[str, Dict[str, float]]: A dictionary with the reaction IDs as keys, and dictionaries as values which,
            in turn, contain the ΔG'° of a reaction under the key 'dG0' and the calculated uncertainty as 'uncertainty'.
    """
    cobra_model = cobra.io.read_sbml_model(sbml_path)

    reaction_dG0s: dict[str, float] = {}
    reaction_dG0_uncertainties: dict[str, float] = {}
    cc = ComponentContribution()
    for reaction_x in cobra_model.reactions:
        reaction: cobra.Reaction = reaction_x

        stop = False
        for exclusion_prefix in exclusion_prefixes:
            if reaction.id.startswith(exclusion_prefix):
                stop = True
        for exclusion_inner_part in exclusion_inner_parts:
            if exclusion_inner_part in reaction.id:
                stop = True
        if stop:
            continue

        stoichiometries: list[float] = []
        compartments: list[str] = []
        identifiers: list[str] = []
        identifier_keys: list[str] = []
        for metabolite_x in reaction.metabolites.keys():
            metabolite: cobra.Metabolite = metabolite_x
            stoichiometries.append(reaction.metabolites[metabolite])
            compartments.append(metabolite.compartment)
            identifier = ""
            for used_identifier in USED_IDENTIFIERS_FOR_EQUILIBRATOR:
                if used_identifier not in metabolite.annotation.keys():
                    continue
                metabolite_identifiers = metabolite.annotation[used_identifier]
                identifier_temp = ""
                if isinstance(metabolite_identifiers, list):
                    identifier_temp = metabolite_identifiers[0]
                elif isinstance(metabolite_identifiers, str):
                    identifier_temp = metabolite_identifiers
                if used_identifier == "inchi":
                    compound = cc.get_compound_by_inchi(identifier_temp)
                elif used_identifier == "inchi_key":
                    compound_list = cc.search_compound_by_inchi_key(identifier_temp)
                    if len(compound_list) > 0:
                        compound = compound_list[0]
                    else:
                        compound = None
                else:
                    identifier_temp = used_identifier + ":" + identifier_temp
                    compound = cc.get_compound(identifier_temp)
                if compound is not None:
                    identifier_key = used_identifier
                    identifier = identifier_temp
                    break
            if identifier == "":
                break
            identifier_keys.append(identifier_key)
            identifiers.append(identifier)

        if identifier == "":
            print(
                f"ERROR: Metabolite {metabolite_x.id} has no identifier of the given types!"
            )
            print(metabolite_x.annotation)
            continue

        # Check for three cases:
        # 1: Single-compartment reaction
        # 2: Double-compartment reaction
        # 3: Multi-compartment reaction (not possible)
        unique_reaction_compartments = list(set(compartments))
        num_compartments = len(unique_reaction_compartments)
        if num_compartments == 1:
            # Set compartment conditions
            compartment = unique_reaction_compartments[0]
            cc.p_h = Q_(phs[compartment])
            cc.p_mg = Q_(pmgs[compartment])
            cc.ionic_strength = Q_(str(ionic_strengths[compartment]) + "mM")

            # Build together reaction
            reaction_dict: dict[Any, float] = {}
            for i in range(len(stoichiometries)):
                identifier_string = identifiers[i]
                identifier_key = identifier_keys[i]
                stoichiometry = stoichiometries[i]
                if identifier_key == "inchi":
                    compound = cc.get_compound_by_inchi(identifier_string)
                elif identifier_key == "inchi_key":
                    compound = cc.search_compound_by_inchi_key(identifier_string)[0]
                else:
                    compound = cc.get_compound(identifier_string)
                reaction_dict[compound] = stoichiometry
            cc_reaction = Reaction(reaction_dict)

            # Check whether or not the reaction is balanced and...
            if not cc_reaction.is_balanced():
                print(f"INFO: Reaction {reaction.id} is not balanced")
                continue

            standard_dg_prime = cc.standard_dg_prime(cc_reaction)
            uncertainty = standard_dg_prime.error.m_as("kJ⋅mol⁻¹")
            if uncertainty < max_uncertainty:
                dG0 = standard_dg_prime.value.m_as("kJ⋅mol⁻¹")
                reaction_dG0s[reaction.id] = dG0
                if ignore_uncertainty:
                    reaction_dG0_uncertainties[reaction.id] = 0.0
                else:
                    reaction_dG0_uncertainties[reaction.id] = abs(uncertainty)

                print(
                    f"No error with reaction {reaction.id}, ΔG'° succesfully calculated!"
                )
            else:
                print(
                    f"INFO: Reaction {reaction.id} uncertainty is too high with {uncertainty} kJ⋅mol⁻¹; ΔG'° not assigned for this reaction"
                )
        elif num_compartments == 2:
            index_zero = inner_to_outer_compartments.index(
                unique_reaction_compartments[0]
            )
            index_one = inner_to_outer_compartments.index(
                unique_reaction_compartments[1]
            )

            if index_one > index_zero:
                outer_compartment = unique_reaction_compartments[1]
                inner_compartment = unique_reaction_compartments[0]
            else:
                outer_compartment = unique_reaction_compartments[0]
                inner_compartment = unique_reaction_compartments[1]

            ph_inner = Q_(phs[inner_compartment])
            ph_outer = Q_(phs[outer_compartment])
            ionic_strength_inner = Q_(str(ionic_strengths[inner_compartment]) + " mM")
            ionic_strength_outer = Q_(str(ionic_strengths[outer_compartment]) + " mM")
            pmg_inner = Q_(pmgs[inner_compartment])
            pmg_outer = Q_(pmgs[outer_compartment])

            if (inner_compartment, outer_compartment) in potential_differences:
                potential_difference = Q_(
                    str(potential_differences[(inner_compartment, outer_compartment)])
                    + " V"
                )
            elif (outer_compartment, inner_compartment) in potential_differences:
                potential_difference = Q_(
                    str(potential_differences[(outer_compartment, inner_compartment)])
                    + " V"
                )
            else:
                print("ERROR")
                continue

            inner_reaction_dict: dict[str, float] = {}
            outer_reaction_dict: dict[str, float] = {}
            for i in range(len(stoichiometries)):
                key = identifiers[i]
                stoichiometry = stoichiometries[i]
                try:
                    compound_key = cc.get_compound(key)
                except Exception:  # sqlalchemy.orm.exc.MultipleResultsFound
                    print("ERROR")
                    continue

                if compound_key is None:
                    print("NONE in compound")
                    continue

                if compartments[i] == inner_compartment:
                    inner_reaction_dict[compound_key] = stoichiometry
                else:
                    outer_reaction_dict[compound_key] = stoichiometry

            cc_inner_reaction = Reaction(inner_reaction_dict)
            cc_outer_reaction = Reaction(outer_reaction_dict)

            cc.p_h = ph_inner
            cc.ionic_strength = ionic_strength_inner
            cc.p_mg = pmg_inner
            try:
                standard_dg_prime = cc.multicompartmental_standard_dg_prime(
                    cc_inner_reaction,
                    cc_outer_reaction,
                    e_potential_difference=potential_difference,
                    p_h_outer=ph_outer,
                    p_mg_outer=pmg_outer,
                    ionic_strength_outer=ionic_strength_outer,
                )
                uncertainty = standard_dg_prime.error.m_as("kJ⋅mol⁻¹")
                if uncertainty < max_uncertainty:
                    dG0 = standard_dg_prime.value.m_as("kJ⋅mol⁻¹")
                    reaction_dG0s[reaction.id] = dG0
                    if ignore_uncertainty:
                        reaction_dG0_uncertainties[reaction.id] = 0.0
                    else:
                        reaction_dG0_uncertainties[reaction.id] = abs(uncertainty)
            except ValueError:
                print("ERROR: Multi-compartmental reaction is not balanced")
                continue
        else:
            print("ERROR: More than two compartments are not possible")
            continue

    return reaction_dG0s, reaction_dG0_uncertainties



import pytest

from cobrak.constants import (
    ALL_OK_KEY,
    ENZYME_VAR_INFIX,
    OBJECTIVE_VAR_NAME,
    REAC_FWD_SUFFIX,
    REAC_REV_SUFFIX,
    SOLVER_STATUS_KEY,
    TERMINATION_CONDITION_KEY,
)
from cobrak.dataclasses import (
    Enzyme,
    EnzymeReactionData,
    ExtraLinearConstraint,
    Metabolite,
    Model,
    Reaction,
)
from cobrak.utilities import (
    add_statuses_to_optimziation_dict,
    compare_optimization_result_reaction_uses,
    delete_orphaned_metabolites_and_enzymes,
    delete_unused_reactions_in_optimization_dict,
    delete_unused_reactions_in_variability_dict,
    get_active_reacs_from_optimization_dict,
    get_base_id,
    get_base_id_optimzation_result,
    get_cobrak_enzyme_reactions_string,
    get_extra_linear_constraint_string,
    get_full_enzyme_id,
    get_full_enzyme_mw,
    get_fwd_rev_corrected_flux,
    get_metabolites_in_elementary_conservation_relations,
    get_potentially_active_reactions_in_variability_dict,
    get_reaction_enzyme_var_id,
    get_reaction_string,
    get_solver_status_from_pyomo_results,
    get_stoichiometric_matrix,
    get_stoichiometrically_coupled_reactions,
    get_substrate_and_product_exchanges,
    get_termination_condition_from_pyomo_results,
    have_all_unignored_km,
    is_objsense_maximization,
    last_n_elements_equal,
    sort_dict_keys,
)


# Example fixtures and mock objects
@pytest.fixture
def mock_optimization_dict():
    return {
        "R1": 1.0,
        "R2": 2.0,
        "R3": 3.0,
        OBJECTIVE_VAR_NAME: 10.0,
        ALL_OK_KEY: True,
        TERMINATION_CONDITION_KEY: 0.2,
        SOLVER_STATUS_KEY: 0,
    }


@pytest.fixture
def mock_pyomo_results():
    class MockSolverResults:
        solver = type(
            "Solver", (object,), {"status": "ok", "termination_condition": "optimal"}
        )()

    return MockSolverResults()


@pytest.fixture
def mock_model():
    return Model(
        reactions={
            "R1": Reaction(
                stoichiometries={"A": -1, "B": 1},
                min_flux=0,
                max_flux=1000,
                dG0=0.0,
                dG0_uncertainty=0.0,
                enzyme_reaction_data=EnzymeReactionData(
                    identifiers=["E1", "E1B"],
                    k_cat=1.0,
                    k_ms={"A": 0.1},
                    k_is={},
                    k_as={},
                    special_stoichiometries={},
                ),
            ),
            "R2": Reaction(
                stoichiometries={"B": -1, "C": 1},
                min_flux=0,
                max_flux=1000,
                dG0=0.0,
                dG0_uncertainty=0.0,
                enzyme_reaction_data=EnzymeReactionData(
                    identifiers=[
                        "E2",
                    ],
                    k_cat=1.0,
                    k_ms={"B": 0.2},
                    k_is={},
                    k_as={},
                    special_stoichiometries={},
                ),
            ),
        },
        metabolites={
            "A": Metabolite(name="A"),
            "B": Metabolite(name="B"),
            "C": Metabolite(name="C"),
        },
        enzymes={
            "E1": Enzyme(
                molecular_weight=10.0,
            ),
            "E1B": Enzyme(
                molecular_weight=10.0,
            ),
            "E2": Enzyme(
                molecular_weight=1.0,
            ),
        },
        max_prot_pool=1000.0,
        R=8.314,
        T=298.15,
        kinetic_ignored_metabolites=[],
    )


@pytest.fixture
def mock_variability_dict():
    return {
        "R1": (0.0, 10.0),
        "R2": (-5.0, 5.0),
        "R3": (0.0, 0.0),
    }


@pytest.fixture
def mock_datasets():
    return [
        {
            "R1": EnzymeReactionData(
                identifiers=["E1"],
                k_cat=1.0,
                k_ms={"A": 0.1},
                k_is={},
                k_as={},
            ),
            "R2": None,
        },
        {
            "R1": None,
            "R2": EnzymeReactionData(
                identifiers=["E2"],
                k_cat=1.0,
                k_ms={},
                k_is={},
                k_as={},
            ),
        },
    ]


# Test functions
def test_add_statuses_to_optimziation_dict(mock_optimization_dict, mock_pyomo_results):
    result = add_statuses_to_optimziation_dict(
        mock_optimization_dict, mock_pyomo_results
    )
    assert result[SOLVER_STATUS_KEY] == 0
    assert result[TERMINATION_CONDITION_KEY] == 0.2
    assert result[ALL_OK_KEY]


def test_get_base_id():
    assert get_base_id(f"R1{REAC_FWD_SUFFIX}") == "R1"
    assert get_base_id(f"R2{REAC_REV_SUFFIX}") == "R2"
    assert get_base_id("R3") == "R3"


def test_get_base_id_optimzation_result(mock_model, mock_optimization_dict):
    result = get_base_id_optimzation_result(mock_model, mock_optimization_dict)
    assert result["R1"] == 1.0
    assert result["R2"] == 2.0


def test_get_fwd_rev_corrected_flux():
    # Test case where reverse reaction exists and has greater flux, and custom fwd and rev
    result = {"R1_fwd": 10.0, "R1_rev": 15.0}
    assert (
        get_fwd_rev_corrected_flux(
            "R1_fwd", {"R1_fwd", "R1_rev"}, result, fwd_suffix="_fwd", rev_suffix="_rev"
        )
        == 0.0
    )

    # Test case where reverse reaction exists and has lesser flux
    result = {f"R1_{REAC_FWD_SUFFIX}": 20.0, f"R1_{REAC_REV_SUFFIX}": 5.0}
    assert (
        get_fwd_rev_corrected_flux(
            f"R1_{REAC_FWD_SUFFIX}",
            {f"R1_{REAC_FWD_SUFFIX}", f"R1_{REAC_REV_SUFFIX}"},
            result,
        )
        == 15.0
    )

    # Test case where reverse reaction does not exist
    result = {f"R1_{REAC_FWD_SUFFIX}": 30.0}
    assert (
        get_fwd_rev_corrected_flux(
            f"R1_{REAC_FWD_SUFFIX}", {f"R1_{REAC_FWD_SUFFIX}"}, result
        )
        == 30.0
    )

    # Test case where reverse reaction is not usable
    result = {f"R1_{REAC_FWD_SUFFIX}": 25.0, f"R1_{REAC_REV_SUFFIX}": 10.0}
    assert (
        get_fwd_rev_corrected_flux(
            f"R1_{REAC_FWD_SUFFIX}", {f"R1_{REAC_FWD_SUFFIX}"}, result
        )
        == 25.0
    )


def test_get_solver_status_from_pyomo_results(mock_pyomo_results):
    assert get_solver_status_from_pyomo_results(mock_pyomo_results) == 0


def test_get_termination_condition_from_pyomo_results(mock_pyomo_results):
    assert get_termination_condition_from_pyomo_results(mock_pyomo_results) == 0.2


def test_sort_dict_keys():
    input_dict = {"b": 2, "a": 1, "c": 3}
    result = sort_dict_keys(input_dict)
    assert result == {"a": 1, "b": 2, "c": 3}


def test_compare_optimization_result_reaction_uses(mock_model, mock_optimization_dict):
    results = [
        mock_optimization_dict,
        {"R1": 0.5, "R2": 1.5, "R3": 2.5, "objective": 15.0},
    ]
    compare_optimization_result_reaction_uses(mock_model, results)
    # This function prints results, so we can't directly assert the output.
    # You can manually check the printed output or redirect stdout to capture it.


def test_delete_orphaned_metabolites_and_enzymes(mock_model):
    model = delete_orphaned_metabolites_and_enzymes(mock_model)
    assert "A" in model.metabolites
    assert "B" in model.metabolites
    assert "C" in model.metabolites
    assert "E1" in model.enzymes


def test_delete_unused_reactions_in_optimization_dict(
    mock_model, mock_optimization_dict
):
    model = delete_unused_reactions_in_optimization_dict(
        mock_model, mock_optimization_dict
    )
    assert "R1" in model.reactions
    assert "R2" in model.reactions
    assert "R3" not in model.reactions


def test_delete_unused_reactions_in_variability_dict(mock_model, mock_variability_dict):
    model = delete_unused_reactions_in_variability_dict(
        mock_model, mock_variability_dict
    )
    assert "R1" in model.reactions
    assert "R2" in model.reactions
    assert "R3" not in model.reactions


def test_get_active_reacs_from_optimization_dict(mock_model, mock_optimization_dict):
    result = get_active_reacs_from_optimization_dict(mock_model, mock_optimization_dict)
    assert result == ["R1", "R2"]


def test_get_cobrak_enzyme_reactions_string(mock_model):
    result = get_cobrak_enzyme_reactions_string(mock_model, "E1")
    assert result == "R1"


def test_get_reaction_string(mock_model):
    result = get_reaction_string(mock_model, "R1")
    assert result == "-1.0 A \u21d2 1.0 B"


def test_get_extra_linear_constraint_string():
    constraint = ExtraLinearConstraint(
        lower_value=0.0,
        upper_value=10.0,
        stoichiometries={"var1": 1.0, "var2": -1.0},
    )
    result = get_extra_linear_constraint_string(constraint)
    assert result == "0.0 \u2264  + 1.0 var1 - 1.0 var2\u2264 10.0"


def test_get_full_enzyme_id():
    result = get_full_enzyme_id(["E1", "E2"])
    assert result == "E1_AND_E2"


def test_get_full_enzyme_mw(mock_model):
    result = get_full_enzyme_mw(mock_model, mock_model.reactions["R1"])
    assert result == 20.0


def test_get_metabolites_in_elementary_conservation_relations(mock_model):
    result = get_metabolites_in_elementary_conservation_relations(mock_model)
    assert sorted(result) == ["A", "B", "C"]


def test_get_potentially_active_reactions_in_variability_dict(
    mock_model, mock_variability_dict
):
    result = get_potentially_active_reactions_in_variability_dict(
        mock_model, mock_variability_dict
    )
    assert result == ["R1", "R2"]


def test_get_reaction_enzyme_var_id(mock_model):
    result = get_reaction_enzyme_var_id("R1", mock_model.reactions["R1"])
    assert result == f"enzyme_E1_AND_E1B{ENZYME_VAR_INFIX}R1"


def test_get_stoichiometric_matrix(mock_model):
    result = get_stoichiometric_matrix(mock_model)
    assert result == [
        [-1, 0.0],
        [1, -1],
        [0.0, 1],
    ]


def test_get_stoichiometrically_coupled_reactions(mock_model):
    result = get_stoichiometrically_coupled_reactions(mock_model)
    assert result == [["R1", "R2"]]


def test_get_substrate_and_product_exchanges(mock_model):
    result = get_substrate_and_product_exchanges(mock_model)
    assert result == (tuple(), tuple())


def test_have_all_unignored_km(mock_model):
    result = have_all_unignored_km(
        mock_model.reactions["R1"], mock_model.kinetic_ignored_metabolites
    )
    assert not result


def test_is_objsense_maximization():
    assert is_objsense_maximization(1)
    assert not is_objsense_maximization(-1)


def test_last_n_elements_equal():
    # Test cases where the last n elements are equal
    assert last_n_elements_equal([1, 2, 3, 4, 4, 4], 3)
    assert last_n_elements_equal(["a", "b", "b", "b"], 3)

    # Test cases where the last n elements are not equal
    assert not last_n_elements_equal([1, 2, 3, 4, 5, 6], 3)
    assert not last_n_elements_equal(["a", "b", "c", "d"], 2)

    # Test cases with fewer elements than n
    assert not last_n_elements_equal([1, 2], 3)
    assert not last_n_elements_equal([], 1)

    # Edge case: n = 0 (should always return True)
    assert last_n_elements_equal([1, 2, 3], 0)
    assert last_n_elements_equal([], 0)
